"""This module differences two CRDS reference or mapping files on the local
system.   It supports specification of the files using only the basenames or
a full path.   Currently it operates on mapping, FITS, or text files.
"""
import os
import sys
from collections import defaultdict
import tempfile
import json
import pprint
import re

# ============================================================================

# from astropy.io.fits import FITSDiff     # deferred

# ============================================================================

import crds
from crds.core import config, log, pysh, utils, rmap
from crds.core import cmdline, naming
from crds import rowdiff, sync

# ============================================================================

__all__ = [
    "DiffScript",
    "difference",
    "get_affected",
    "get_added_references",
    "get_deleted_references",
    "diff_action",
    "diff_replace_old_new",
    "mapping_check_diffs",
    "mapping_diffs",
]

# ============================================================================

def mapping_diffs(old_file, new_file, *args, **keys):
    """Return the logical differences between CRDS mappings named `old_file`
    and `new_file`.

    IFF include_header_diffs,  include differences in mapping headers.   Some are "boring", e.g. sha1sum or name.

    IFF recurse_added_deleted,  include difference tuples for all nested adds and deletes whenever a higher level
        mapping is added or deleted.   Else, only include the higher level mapping,  not contained files.
    """
    observatory = keys.pop("observatory", None)
    differ = MappingDifferencer(observatory, old_file, new_file, *args, **keys)
    return differ.mapping_diffs()

def get_affected(old_pmap, new_pmap, *args, **keys):
    """Examine the diffs between `old_pmap` and `new_pmap` and return sorted lists of affected instruments and types.

    Returns { affected_instrument : { affected_type, ... } }
    """
    observatory = keys.pop("observatory", crds.get_pickled_mapping(old_pmap).observatory)  # reviewed
    differ = MappingDifferencer(observatory, old_pmap, new_pmap, *args, **keys)
    return differ.get_affected()

# ==============================================================================================================

def difference(observatory, old_file, new_file, *args, **keys):
    """Difference different kinds of CRDS files (mappings, FITS references, etc.)
    named `old_file` and `new_file` against one another and print out the results
    on stdout.

    Returns:

    0 no differences
    1 some differences
    2 errors in subshells

    """
    filetype = config.filetype(old_file)
    differs = {
        "mapping" : MappingDifferencer,
        "asdf" : AsdfDifferencer,
        "fits" : FitsDifferencer,
        "text" : TextDifferencer,
        "yaml" : TextDifferencer,
        "json" : JsonDifferencer,
        }
    differ_class = differs.get(filetype, None)
    if differ_class is None:
        log.warning("Cannot difference file of type", repr(filetype), ":", repr(old_file), repr(new_file))
        status = 2   #  arguably, this should be an error not a warning.  wary of changing now.
    else:
        differ = differ_class(observatory, old_file, new_file, *args, **keys)
        status = differ.difference()
    return status

# ============================================================================

def decolorize(output):
    """Remove ANSI color codes,  particularly for doc-testing."""
    return re.sub(r"\x1B\[([0-9]{1,2}(;[0-9]{1,2})?)?[m|K]", "", output)

# ==============================================================================================================

class Differencer:
    """This baseclass provides a standard set of (largely optional) parameters to subclasses.

    observatory           str  nominally hst or jwst
    old_file              str  name of first/old file in difference
    new_file              str  name of second/new file in difference
    primitive_diffs       bool difference replaced reference files using fitsdiff
    check_diffs           bool check for name reversions, file deletions, rule changes, class changes, and parameter changes
    check_references      bool check for reference additions or deletions, regardless of matching criteria
    mapping_text_diffs    bool print out UNIX text diffs for mapping changes
    include_header_diffs  bool include header changes in logical diff output for mappings
    hide_boring_diffs     bool hide boring header changes in logical diff output (name, sha1sum, derived_from, etc.)
    recurse_added_deleted bool recursively include all files of added or deleted mapping as also added or deleted
    lowest_mapping_only   bool in each logical diff output,  only show the lowest level mapping,  not the full context traversal
    remove_paths          bool remove paths from mapping names in logical diffs
    """

    def __init__(self, observatory, old_file, new_file, primitive_diffs=False, check_diffs=False, mapping_text_diffs=False,
                 include_header_diffs=False, hide_boring_diffs=False, recurse_added_deleted=False,
                 lowest_mapping_only=False, remove_paths=False, squash_tuples=False, check_references=False,
                 cache1=None, cache2=None):
        self.observatory = observatory
        self.old_file = old_file
        self.new_file = new_file
        self.primitive_diffs = primitive_diffs
        self.check_diffs = check_diffs
        self.check_references = check_references
        self.mapping_text_diffs = mapping_text_diffs
        self.include_header_diffs = include_header_diffs
        self.hide_boring_diffs = hide_boring_diffs
        self.recurse_added_deleted = recurse_added_deleted
        self.lowest_mapping_only = lowest_mapping_only
        self.remove_paths = remove_paths
        self.squash_tuples = squash_tuples
        self.cache1 = cache1
        self.cache2 = cache2
        self.mappings_cache1 = os.path.join(self.cache1, "mappings", self.observatory) if cache1 else None
        self.mappings_cache2 = os.path.join(self.cache2, "mappings", self.observatory) if cache2 else None

    def locate_file(self, filename, cache=None):
        """Return the full path for `filename` implementing default CRDS file cache
        location behavior,  and verifying that the resulting path is safe.  If cache
        is defined,  override CRDS_PATH and any path included in `filename`.
        """
        if cache is not None:
            os.environ["CRDS_PATH"] = cache
            filename = os.path.basename(filename)
        path = config.locate_file(filename, self.observatory)
        config.check_path(path)  # check_path returns abspath,  bad for listsings.
        return path

    def locate_file1(self, filename):
        """Locate `filename` inside CRDS cache,  potentially redefining cache as self.cache1."""
        return self.locate_file(filename, self.cache1)

    def locate_file2(self, filename):
        """Locate `filename` inside CRDS cache,  potentially redefining cache as self.cache1."""
        return self.locate_file(filename, self.cache2)

    def difference(self):
        """Run diff() method on `old_file` and `new_file` and print() output.

        Returns the output status and

        Returns:
        0 no differences
        1 some differences
        2 errors

        NOTE:  often overridden
        """
        status, out_err = self.diff()
        print(out_err, end="")
        return status

    def diff(self):
        """Placeholder diff() method for abstract class."""
        raise NotImplementedError("Differencer is an abstract class.")
# ==============================================================================================================

class MappingDifferencer(Differencer):
    """This class recursively differences a mapping hierarchy."""

    def __init__(self, *args, **keys):
        super(MappingDifferencer, self).__init__(*args, **keys)
        assert config.is_mapping(self.old_file), \
            "File " + repr(self.old_file) + " is not a CRDS mapping."
        assert config.is_mapping(self.new_file), \
            "File " + repr(self.new_file) + " is not a CRDS mapping."
        assert os.path.splitext(self.old_file)[-1] == os.path.splitext(self.new_file)[-1], \
            "Files " + repr(self.old_file) + " and " + repr(self.new_file) + \
            " are not the same kind of CRDS mapping:  .pmap, .imap, .rmap"

    def difference(self):
        """Print the logical differences between CRDS mappings named `old_file`
        and `new_file`.

        IFF primitive_differences,  recursively difference any replaced files found
        in the top level logical differences.

        IFF check_diffs, issue warnings about critical differences.   See
        mapping_check_diffs().
        """
        differences = self.mapping_diffs()
        if self.mapping_text_diffs:   # only banner when there's two kinds to differentiate
            log.write("="*20, "logical differences",  repr(self.old_file), "vs.", repr(self.new_file), "="*20)
        if self.hide_boring_diffs:
            differences = remove_boring(differences)
        for diff in sorted(differences):
            diff1 = unquote_diff(diff)
            if self.primitive_diffs:
                log.write("="*80)
            diff2 = simplify_to_lowest_mapping(diff1) if self.lowest_mapping_only else diff1
            diff2 = remove_diff_paths(diff2) if self.remove_paths else diff2
            if not self.squash_tuples:
                log.write(diff2)
            else:
                log.write(self.squash_diff_tuples(diff2))
            if self.primitive_diffs and "header" not in diff_action(diff1):
                # XXXX fragile, coordinate with selector.py and rmap.py
                if "replaced" in diff[-1]:
                    old, new = diff_replace_old_new(diff)
                    if old and new:
                        from crds import diff
                        diff.difference(self.observatory, old, new, primitive_diffs=self.primitive_diffs,
                                        recurse_added_deleted=self.recurse_added_deleted)
        pairs = sorted(set(mapping_pairs(differences) +  [(self.old_file, self.new_file)]))
        for (old, new) in pairs:
            if self.mapping_text_diffs:
                log.write("="*20, "text difference", repr(old), "vs.", repr(new), "="*20)
                text_difference(self.observatory, old, new)
                log.write("="*80)
            if self.check_references:
                mapping_check_references(new, old)

        if self.check_diffs:
            mapping_check_diffs_core(differences)

        return 1 if differences else 0

    def squash_diff_tuples(self, diff2):
        """Change notation of diff tuples into -- separated components for readability."""
        return " -- ".join([" ".join(diff) if isinstance(diff, tuple) else diff for diff in diff2])

    def mapping_diffs(self):
        """Return the logical differences between CRDS mappings named `old_file`
        and `new_file`.

        IFF include_header_diffs,  include differences in mapping headers.   Some are "boring", e.g. sha1sum or name.

        IFF recurse_added_deleted,  include difference tuples for all nested adds and deletes whenever a higher level
        mapping is added or deleted.   Else, only include the higher level mapping,  not contained files.

        """
        # At this time,  the fetch_mapping path parameter appears to exist only to thwart CRDS mapping caching.
        old_map = rmap.fetch_mapping(self.locate_file1(self.old_file), ignore_checksum=True, path=self.mappings_cache1)
        new_map = rmap.fetch_mapping(self.locate_file2(self.new_file), ignore_checksum=True, path=self.mappings_cache2)
        differences = old_map.difference(new_map, include_header_diffs=self.include_header_diffs,
                                         recurse_added_deleted=self.recurse_added_deleted)
        return differences

    def get_affected(self):
        """Examine the diffs between `old_pmap` and `new_pmap` and return sorted lists of affected instruments and types.

        Returns { affected_instrument : { affected_type, ... } }
        """
        instrs = defaultdict(set)
        diffs = self.mapping_diffs()
        diffs = remove_boring(diffs)
        for diff in diffs:
            for step in diff:
                # Walking down the diff steps 1-by-1 eventually hits an rmap comparison which
                # will define both instrument and type.  pmaps and imaps leave at least one blank.
                if len(step) == 2 and config.is_mapping(step[0]):
                    instrument, filekind = utils.get_file_properties(self.observatory, step[0])
                # This is inefficient since diff doesn't vary by step,  but set logic cleans up the redundancy
                # New rmaps imply reprocessing the entire type.
                elif isinstance(diff[-1],str) and diff[-1].startswith(("added","deleted")) and \
                        diff[-1].endswith(".rmap'"):
                    rmap_name = diff[-1].split()[-1].replace("'","")
                    rmapping = rmap.fetch_mapping(rmap_name, ignore_checksum=True)
                    instrument, filekind = rmapping.instrument, rmapping.filekind
                if instrument.strip() and filekind.strip():
                    if filekind not in instrs[instrument]:
                        log.verbose("Affected", (instrument, filekind), "based on diff", diff, verbosity=20)
                        instrs[instrument].add(filekind)
        return { key:list(val) for (key, val) in instrs.items() }

    def header_modified(self):
        """Return true IFF there were changes in an rmap header."""
        return self._find_diff_str("header")

    def files_deleted(self):
        """Return True IFF files were deleted in an rmap."""
        return self._find_diff_str("delete")

    def _find_diff_str(self, diff_str):
        """Return True IFF `diff_str` is in some rmap diff."""
        diffs = self.mapping_diffs()
        diffs = remove_boring(diffs)
        for diff in diffs:
            for step in diff:
                if len(step) == 2 and config.is_mapping(step[0]):
                    if diff_str in diff_action(diff):
                        log.verbose("Found", repr(diff_str), "diff between", repr(step[0:1]))
                        return True
        return False


# ==============================================================================================================

class FitsDifferencer(Differencer):

    """Differences FITS format files including row differences."""

    def __init__(self, *args, **keys):
        super(FitsDifferencer, self).__init__(*args, **keys)
        assert self.old_file.endswith(".fits"), \
            "File " + repr(self.old_file) + " is not a FITS file."
        assert self.new_file.endswith(".fits"), \
            "File " + repr(self.new_file) + " is not a FITS file."

    def difference(self):
        """Run fitsdiff on files named `old_file` and `new_file`.

        Returns:

        0 no differences
        1 some differences
        """
        loc_old_file = self.locate_file1(self.old_file)
        loc_new_file = self.locate_file2(self.new_file)

        # Do the standard diff.
        from astropy.io.fits import FITSDiff           # deferred
        fdiff = FITSDiff(loc_old_file, loc_new_file)

        # Do the diff by rows.
        rdiff = rowdiff.RowDiff(loc_old_file, loc_new_file)

        if not fdiff.identical:
            fdiff.report(fileobj=sys.stdout)
            print('\n', rdiff)

        return 0 if fdiff.identical else 1

# ==============================================================================================================

class TextDifferencer(Differencer):
    """Run UNIX diff on two text files named `old_file` and `new_file`."""

    def __init__(self, *args, **keys):
        """Initialize TextDifferencer validating identical extensions for both files."""
        super(TextDifferencer, self).__init__(*args, **keys)
        assert os.path.splitext(self.old_file)[-1] == os.path.splitext(self.new_file)[-1], \
            "Files " + repr(self.old_file) + " and " + repr(self.new_file) + " are of different types."

    def diff(self):
        """Returns the diff status and combined output from stdout and stderr of the diff command."""
        loc_old_file = self.locate_file1(self.old_file)
        loc_new_file = self.locate_file2(self.new_file)
        status, out_err = pysh.status_out_err("diff -b -c ${loc_old_file} ${loc_new_file}", raise_on_error=False)   # secure
        return status, out_err

# ==============================================================================================================

class AsdfDifferencer(Differencer):
    """Run UNIX diff on two text files named `old_file` and `new_file`."""

    def __init__(self, *args, **keys):
        """Initialize TextDifferencer validating identical extensions for both files."""
        super(AsdfDifferencer, self).__init__(*args, **keys)
        assert os.path.splitext(self.old_file)[-1] == os.path.splitext(self.new_file)[-1], \
            "Files " + repr(self.old_file) + " and " + repr(self.new_file) + " are of different types."

    def diff(self):
        """Returns the diff status and combined output from stdout and stderr of the diff command."""
        loc_old_file = self.locate_file1(self.old_file)
        loc_new_file = self.locate_file2(self.new_file)
        status, out_err = pysh.status_out_err("asdftool diff ${loc_old_file} ${loc_new_file}", raise_on_error=False)   # secure
        if not status and len(out_err) != 0:
            # convert asdftool "no errors" to diff-style "diffs exist"
            status = 1
        if len(out_err) == 0:
            out_err = ""  # otherwise b''
        else:
            # asdftool colorizes diffs using ANSI color escape sequences.
            out_err = decolorize(out_err)
        return status, out_err

# ==============================================================================================================

class JsonDifferencer(TextDifferencer):
    """Run UNIX diff on two text files named `old_file` and `new_file`."""

    def __init__(self, *args, **keys):
        """Initialize JsonDifferencer capturing optional keyword parameter pretty_print=True."""
        super(JsonDifferencer, self).__init__(*args, **keys)
        self.pretty_print = keys.pop("pretty_print", True)
        self.remove_files = []

    def locate_file(self, filename):
        """Create a temporary file based on source filename source name IFF self.pretty_print."""
        if not self.pretty_print:
            return super(JsonDifferencer, self).locate_file(filename)
        else:
            source_path = config.check_path(filename)
            temp_file = tempfile.NamedTemporaryFile(delete=False)
            with open(source_path) as source_file:
                source_json = json.load(source_file)
            source_pp = pprint.pformat(utils.fix_json_strings(source_json))
            temp_file.write(source_pp)
        self.remove_files += [temp_file.name]
        return temp_file.name

    def difference(self):
        """Run UNIX diff on two json files named `old_file` and `new_file`.
        If self.pretty_print is True,  reformat the files as temporaries

        Returns:
        0 no differences
        1 some differences
        2 errors
        """
        result, out_err = super(JsonDifferencer, self).diff()
        out_err = out_err.replace(self.remove_files[0], self.old_file)
        out_err = out_err.replace(self.remove_files[1], self.new_file)
        print(out_err, end="")
        if self.pretty_print:
            for fname in self.remove_files:
                with log.verbose_warning_on_exception("Failed removing", repr(fname)):
                    os.remove(fname)
        return result

# ============================================================================

def simplify_to_lowest_mapping(diff):
    """For diffs which affect nested mappings,  pop-off the path of the higher level mappings
    and show only the name of the leaf mapping.

    >>> simplify_to_lowest_mapping((('jwst_0063.pmap', 'jwst_0074.pmap'), ('jwst_nirspec_0024.imap', 'jwst_nirspec_0027.imap'),
    ...                            ('jwst_nirspec_saturation_0001.rmap', 'jwst_nirspec_saturation_0002.rmap'),
    ...                            "header replaced 'parkey' = (('META.INSTRUMENT.DETECTOR', 'META.INSTRUMENT.FILTER'),) with (('META.INSTRUMENT.DETECTOR',),)"))
    (('jwst_nirspec_saturation_0001.rmap', 'jwst_nirspec_saturation_0002.rmap'), "header replaced 'parkey' = (('META.INSTRUMENT.DETECTOR', 'META.INSTRUMENT.FILTER'),) with (('META.INSTRUMENT.DETECTOR',),)")
    """
    log.verbose("simplify_to_lowest_mapping:", diff)
    lowest_ranked_mapping = None
    for tup in diff:
        if isinstance(tup, tuple) and  len(tup) == 2 and isinstance(tup[0], str) and tup[0].endswith("map"):
            lowest_ranked_mapping = tup[0]
    simplified = []
    for tup in diff:
        if len(tup) == 2 and isinstance(tup[0], str) and tup[0].endswith("map"):
            if tup[0] != lowest_ranked_mapping:
                continue
        simplified.append(tup)
    return tuple(simplified)

def remove_diff_paths(diff):
    """Remove the paths from leading diff file names."""
    simplified = []
    for tup in diff:
        if len(tup) == 2 and isinstance(tup[0], str) and tup[0].endswith("map"):
            tup = tuple([os.path.basename(t) for t in tup])
        simplified.append(tup)
    return tuple(simplified)

def mapping_pairs(differences):
    """Return the sorted list of all mapping tuples found in differences."""
    pairs = set()
    for diff in differences:
        for pair in diff:
            if len(pair) == 2 and config.is_mapping(pair[0]):
                pairs.add(pair)
    return sorted(pairs)

def unquote_diff(diff):
    """Remove repr str quoting in `diff` tuple,  don't change header diffs."""
    return diff[:-1] + (diff[-1].replace("'",""),) if "header" not in diff[-1] else diff

def unquote(name):
    """Remove string quotes from simple `name` repr."""
    return name.replace("'","").replace('"','')

# XXXX fragile,  coordinate with selector.py and rmap.py
def diff_replace_old_new(diff):
    """Return the (old, new) filenames from difference tuple `diff`."""
    match = re.search(r"replaced '(.+)' with '(.+)'", diff[-1])
    if match:
        return match.group(1), match.group(2)
    else:
        return None, None

# XXXX fragile,  coordinate with selector.py and rmap.py
def diff_added_new(diff):
    """Return the (old, new) filenames from difference tuple `diff`."""
    new = diff[-1].split()[-1]
    return unquote(new)

# XXXX fragile,  coordinate with selector.py and rmap.py
def diff_action(diff):
    """Return 'add', 'replace', or 'delete' based on action represented by
    difference tuple `d`.   Append "_rule" if the change is a Selector.
    """
    if "replace" in diff[-1]:
        result = "replace"
    elif "add" in diff[-1]:
        result = "add"
    elif "delete" in diff[-1]:
        result = "delete"
    elif "different classes" in diff[-1]:
        result = "class_difference"
    elif "different parameter" in diff[-1]:
        result = "parkey_difference"
    else:
        raise ValueError("Bad difference action: "  + repr(diff))
    if "rule" in diff[-1]:
        result += "_rule"
    elif "header" in diff[-1]:
        result += "_header"
    return result
# ============================================================================

def mapping_affected_modes(old_file, new_file, include_header_diffs=True):
    """Return a sorted set of flat tuples describing the matching parameters affected by the differences
    between old_file and new_file.
    """
    affected = defaultdict(int)
    differences = mapping_diffs(old_file, new_file, include_header_diffs=include_header_diffs)
    for diff in differences:
        mode = affected_mode(diff)
        if mode is not None:
            affected[mode] += 1
    return [ tup + (("DIFF_COUNT", str(affected[tup])),) for tup in sorted(affected) ]

DEFAULT_EXCLUDED_PARAMETERS = ("DATE-OBS", "TIME-OBS",
                               "META.OBSERVATION.DATE", "META.OBSERVATION.TIME",
                               "META_OBSERVATION_DATE", "META_OBSERVATION_TIME",
                               "DIFFERENCE", "INSTRUME", "REFTYPE")

BORING_VARS = ["NAME", "DERIVED_FROM", "SHA1SUM", "ROW_KEYS"]  # XXX ROW_KEYS obsolete

def affected_mode(diff, excluded_parameters=DEFAULT_EXCLUDED_PARAMETERS):
    """Return a list of parameter items which characterize the effect of a difference in
    terms of instrument, type, and instrument mode.
    """
    instrument = filekind = None
    affected = []
    flat = diff.flat
    for (i, val) in enumerate(flat):
        var = flat.parameter_names[i]
        if var == "PipelineContext":
            pass
        elif var == "InstrumentContext":
            instrument = diff.instrument
        elif var == "ReferenceMapping":
            instrument = diff.instrument
            filekind = diff.filekind
        elif var not in excluded_parameters:
            affected.append((var, val))
        elif var == "DIFFERENCE" and "header" in val:
            for boring in BORING_VARS:
                if "REPLACED " + repr(boring) in val.upper():
                    break
            else:
                affected.append((var, val))
    if not affected:
        return None
    if filekind:
        affected = [("REFTYPE", filekind.upper())] + affected
    if instrument:
        affected = [("INSTRUMENT", instrument.upper())] + affected
    return tuple(affected)

def format_affected_mode(mode):
    """Format an affected mode as a string."""
    return " ".join(["=".join([item[0], repr(item[1])]) for item in mode])

def remove_boring(diffs):
    """Remove routine differences from a list of diff tuples."""
    result = diffs[:]
    for diff in diffs:
        for var in BORING_VARS:
            if var.lower() in diff[-1]:
                result.remove(diff)
    return result

# ============================================================================

def mapping_check_references(mapping, derived_from):
    """Regardless of matching criteria,  do a simple check listing added or deleted
    references as appropritate.
    """
    mapping = rmap.asmapping(mapping, cached="readonly")
    derived_from = rmap.asmapping(derived_from, cached="readonly")
    old_refs = set(derived_from.reference_names())
    new_refs = set(mapping.reference_names())
    if old_refs - new_refs:
        log.warning("Deleted references for", repr(derived_from.filename), "and", repr(mapping.filename), "=",
                 list(old_refs - new_refs))
    if new_refs - old_refs:
        log.warning("Added references for", repr(derived_from.filename), "and", repr(mapping.filename), "=",
                 list(new_refs - old_refs))

def mapping_check_diffs(mapping, derived_from):
    """Issue warnings for *deletions* in self relative to parent derived_from
    mapping.  Issue warnings for *reversions*,  defined as replacements which
    where the replacement is older than the original,  as defined by the names.

    This is intended to check for missing modes and for inadvertent reversions
    to earlier versions of files.   For speed and simplicity,  file time order
    is currently determined by the names themselves,  not file contents, file
    system,  or database info.
    """
    mapping = rmap.asmapping(mapping, cached="readonly")
    derived_from = rmap.asmapping(derived_from, cached="readonly")
    log.info("Checking diffs from", repr(derived_from.basename), "to", repr(mapping.basename))
    diffs = derived_from.difference(mapping)
    mapping_check_diffs_core(diffs)

def mapping_check_diffs_core(diffs):
    """Perform the core difference checks on difference tuples `diffs`."""
    categorized = sorted([ (diff_action(d), d) for d in diffs ])
    for action, msg in categorized:
        if "header" in action:
            log.verbose("In", _diff_tail(msg)[:-1], msg[-1])
        elif action == "add":
            log.verbose("In", _diff_tail(msg)[:-1], msg[-1])
        elif "rule" in action:
            log.warning("Rule change at", _diff_tail(msg)[:-1], msg[-1])
        elif action == "replace":
            old_val, new_val = diff_replace_old_new(msg)
            if old_val and new_val:
                old_val, new_val = [x for x in diff_replace_old_new(msg)]
                if naming.newer(new_val, old_val):
                    log.verbose("In", _diff_tail(msg)[:-1], msg[-1])
                else:
                    log.warning("Reversion at", _diff_tail(msg)[:-1], msg[-1])
            else:
                log.warning("Unusual replacement", _diff_tail(msg)[:-1], msg[-1])
        elif action == "delete":
            log.warning("Deletion at", _diff_tail(msg)[:-1], msg[-1])
        elif action == "parkey_difference":
            log.warning("Different lookup parameters", _diff_tail(msg)[:-1], msg[-1])
        elif action == "class_difference":
            log.warning("Different classes at", _diff_tail(msg)[:-1], msg[-1])
        else:
            raise ValueError("Unexpected difference action:", action)

def _diff_tail(msg):
    """`msg` is an arbitrary length difference "path",  which could
    be coming from any part of the mapping hierarchy and ending in any kind of
    selector tree.   The last item is always the change message: add, replace,
    delete <blah>.  The next to last should always be a selector key of some kind.
    Back up from there to find the first mapping tuple.
    """
    tail = []
    for part in msg[::-1]:
        if isinstance(part, tuple) and len(part) == 2 and isinstance(part[0], str) and part[0].endswith("map"):
            tail.append(part[1])
            break
        else:
            tail.append(part)
    return tuple(reversed(tail))

# ============================================================================


def fits_difference(*args, **keys):
    """Difference two FITS files with parameters specified as Differencer class."""
    differ = FitsDifferencer(*args, **keys)
    return differ.difference()

def text_difference(*args, **keys):
    """Run UNIX diff on two text files named `old_file` and `new_file`.  Parameters
    are specified as Differencer class.

    Returns:
      0 no differences
      1 some differences
      2 errors
    """
    differ = TextDifferencer(*args, **keys)
    return differ.difference()

# ==============================================================================================================

def get_added_references(old_pmap, new_pmap, cached=True):
    """Return the list of references from `new_pmap` which were not in `old_pmap`."""
    old_pmap = rmap.asmapping(old_pmap, cached=cached)
    new_pmap = rmap.asmapping(new_pmap, cached=cached)
    return sorted(list(set(new_pmap.reference_names()) - set(old_pmap.reference_names())))

def get_deleted_references(old_pmap, new_pmap, cached=True):
    """Return the list of references from `new_pmap` which were not in `old_pmap`."""
    old_pmap = rmap.asmapping(old_pmap, cached=cached)
    new_pmap = rmap.asmapping(new_pmap, cached=cached)
    return sorted(list(set(old_pmap.reference_names()) - set(new_pmap.reference_names())))

def get_updated_files(context1, context2):
    """Return the sorted list of files names which are in `context2` (or any intermediate context)
    but not in `context1`.   context2 > context1.
    """
    extension1 = os.path.splitext(context1)[1]
    extension2 = os.path.splitext(context2)[1]
    assert extension1 == extension2, "Only compare mappings of same type/extension."
    old_map = crds.get_cached_mapping(context1)
    old_files = set(old_map.mapping_names() + old_map.reference_names())
    all_mappings = rmap.list_mappings("*"+extension1, old_map.observatory)
    updated = set()
    context1, context2 = os.path.basename(context1), os.path.basename(context2)
    for new in all_mappings:
        new = os.path.basename(new)
        if context1 < new <= context2:
            new_map = crds.get_cached_mapping(new)
            updated |= set(new_map.mapping_names() + new_map.reference_names())
    return sorted(list(updated - old_files))

# ==============================================================================================================

class DiffScript(cmdline.Script):
    """Python command line script to difference mappings or references."""

    description = """Difference CRDS mapping or reference files."""

    epilog = """Reference files are nominally differenced using FITS-diff or diff.

Mapping files are differenced using CRDS machinery to recursively compare too mappings and
their sub-mappings.

Differencing two mappings will find all the logical differences between the two contexts
and any nested mappings.

By specifying --mapping-text-diffs,  UNIX diff will be run on mapping files in addition to
CRDS logical diffs.   (Duplicate/overlapping cases can be collapsed by the default mapping loader
and hence missed in logical differences,  but are revealed by text differences.   crds.certify
can also detect duplicate entries,  but at the cost of mapping load speed.)

By specifying --primitive-diffs,  FITS diff will be run on all references which are replaced
in the logical differences between two mappings.

For example:

    % crds diff hst_0001.pmap  hst_0005.pmap  --brief

Or to include UNIX diff style diffs of rules as well:

    % crds diff hst_0001.pmap  hst_0005.pmap  --brief --mapping-text-diffs

Or for more recursive diffs,  including reference file diffs similar to fitsdiff:

    % crds diff hst_0001.pmap  hst_0005.pmap  --brief --primitive-diffs

NOTE: mapping logical differences (the default) do not compare CRDS mapping headers,  use
--brief or read --help about other switch options.

NOTE: mapping logical differences do not normally include nested files which are implicitly
added or deleted by a hgher level mapping change.  See --recurse-added-deleted to include those.

RETURNS: crds.diff has "non-standard" exit status similar to UNIX diff:
     0 no differences
     1 some differences
     2 errors or warnings

Differencing two sets of rules (withing the same cache) with simplified output:

    % crds diff jwst_0080.pmap jwst_0081.pmap --brief --squash-tuples
    jwst_miri_regions_0004.rmap jwst_miri_regions_0005.rmap -- MIRIFUSHORT 12 SHORT N/A -- added Match rule for jwst_miri_regions_0006.fits
    jwst_miri_0048.imap jwst_miri_0049.imap -- regions -- replaced jwst_miri_regions_0004.rmap with jwst_miri_regions_0005.rmap
    jwst_0080.pmap jwst_0081.pmap -- miri -- replaced jwst_miri_0048.imap with jwst_miri_0049.imap

Differencing two sets of rules (from two different pre-synced caches, e.g. from TEST and OPS)  rules only:

    # First make sure two rules caches are up to date by syncing
    % .... sync cache #1 using crds.sync and server #1
    % .... sync cache #2 using crds.sync and server #2

    % crds diff --cache1=/Users/fred/crds_cache_test --cache2=/Users/fred/crds_cache_ops hst_0382.pmap hst_0422.pmap  -F -Q
    ...

    --cache1 and --cache2 are required as a pair.

    This is a direct approach for recursively differencing a version of rules from the TEST pipeline with rules from the OPS pipeline.
    Not all differencing modes work for this feature,  it's intended only for comparing rules,  doesn't support direct sync'ing, etc.
    A key point is that different rules files in TEST and OPS can have the same name.

    NOTE:  For cache-to-cache differences,  --sync-files is not supported.  Compared caches are assumed to pre-exist.

Mutually Exclusive Modes

    The following switches define diffencing modes which cannot be run together:

    --print-new-files
    --print-all-new-files
    --print-affected-instruments
    --print-affected-types
    --print-affected-modes
    --cache1 (or --cache2)
    (default differencing with none of the above)

    --sync-files and --cache1 / --cache2 are also mutally exclusive

    """
    def __init__(self, *args, **keys):
        super(DiffScript, self).__init__(*args, **keys)
        self.old_file = None
        self.new_file = None

    locate_file = cmdline.Script.locate_file_outside_cache

    def add_args(self):
        """Add diff-specific command line parameters."""
        self.add_argument("old_file",  help="Prior file of difference.""")
        self.add_argument("new_file",  help="New file of difference.""")

        self.add_argument("-P", "--primitive-diffs", dest="primitive_diffs",
            help="Fitsdiff replaced reference files when diffing mappings.",
            action="store_true")

        self.add_argument("-T", "--mapping-text-diffs",  dest="mapping_text_diffs", action="store_true",
            help="In addition to CRDS mapping logical differences,  run UNIX context diff for mappings.")
        self.add_argument("-i", "--include-header-diffs", dest="include_header_diffs", action="store_true",
            help="Include mapping header differences in logical diffs: rmap_relevance, sha1sum, derived_from, etc.")
        self.add_argument("-B", "--hide-boring-diffs", dest="hide_boring_diffs", action="store_true",
            help="Remove boiler-plate header differences in logical diffs: sha1sum, derived_from, etc.")
        self.add_argument("-U", "--recurse-added-deleted",  dest="recurse_added_deleted", action="store_true",
            help="When a mapping is added or deleted, include all nested files as also added or deleted.  Else only top mapping change listed.")

        self.add_argument("-K", "--check-diffs", dest="check_diffs", action="store_true",
            help="Issue warnings about new rules, deletions, or reversions.")
        self.add_argument("--check-references", dest="check_references", action="store_true",
            help="Issue warnings if references are added to or deleted from either mapping.")

        group = self.get_exclusive_arg_group(required=False)
        group.add_argument("-N", "--print-new-files", dest="print_new_files", action="store_true",
            help="Rather than printing diffs for mappings,  print the names of new or replacement files.  Excludes intermediaries.")
        group.add_argument("-A", "--print-all-new-files", dest="print_all_new_files", action="store_true",
            help="Print the names of every new or replacement file in diffs between old and new.  Includes intermediaries.")
        group.add_argument("--print-affected-instruments", dest="print_affected_instruments", action="store_true",
            help="Print out the names of instruments which appear in diffs,  rather than diffs.")
        group.add_argument("--print-affected-types", dest="print_affected_types", action="store_true",
            help="Print out the names of instruments and types which appear in diffs,  rather than diffs.")
        group.add_argument("--print-affected-modes", dest="print_affected_modes", action="store_true",
            help="Print out the names of instruments, types, and matching parameters,  rather than diffs.")
        group.add_argument("--cache1",  dest="cache1", default=None,
            help="CRDS_PATH for the first cache in a cache-to-cache difference.  Mappings only.""")

        # Can't exclude both cache1 and cache2
        # --sync-files exludes --cache1 and --cache2 but not all the others,  --cache1 and --cache2 required together
        group2 = self.get_exclusive_arg_group(required=False)
        group2.add_argument("--cache2",  dest="cache2", default=None,
            help="CRDS_PATH for the second cache in a cache-to-cache difference.  Mappings only.""")
        group2.add_argument("--sync-files", dest="sync_files", action="store_true",
            help="Fetch any missing files needed for the requested difference from the CRDS server.")

        self.add_argument("-L", "--lowest-mapping-only", dest="lowest_mapping_only", action="store_true",
            help="Only include the name of the leaf mapping being diffed,  not all anscestor mappings.")
        self.add_argument("-E", "--remove-paths", dest="remove_paths", action="store_true",
            help="Remove path names from files in output.")
        self.add_argument("-Q", "--squash-tuples", dest="squash_tuples", action="store_true",
            help="Simplify formatting of difference results (remove tuple notations)")
        self.add_argument("-F", "--brief", dest="brief", action="store_true",
            help="Switch alias for --lowest-mapping-only --remove-paths --hide-boring-diffs --include-headers")

    def main(self):
        """Perform the differencing."""
        self.args.files = [ self.args.old_file, self.args.new_file ]   # for defining self.observatory
        self.old_file = self.locate_file(self.args.old_file)
        self.new_file = self.locate_file(self.args.new_file)
        if self.args.brief:
            self.args.lowest_mapping_only = True
            self.args.remove_paths = True
            self.args.hide_boring_diffs = True
            self.args.include_header_diffs = True
        if self.args.sync_files:
            assert not (self.args.cache1 or self.args.cache2), \
                "--sync-files is not compatible with cache-to-cache differences."
            if self.args.print_all_new_files:
                serial_old = naming.newstyle_serial(self.old_file)
                serial_new = naming.newstyle_serial(self.new_file) + 1
                if None not in [serial_old, serial_new]:
                    errs = sync.SyncScript("crds.sync --range {0}:{1}".format(serial_old, serial_new))()
                    assert not errs, "Errors occurred while syncing all rules to CRDS cache."
                else:
                    log.warning("Cannot sync non-standard mapping names,  results may be incomplete.")
            else:
                self.sync_files([self.old_file, self.new_file])
        elif self.args.print_all_new_files:
            log.warning("--print-all-new-files requires a complete set of rules.  suggest --sync-files.")

        # self.args.files = [ self.old_file, self.new_file ]   # for defining self.observatory

        assert (self.args.cache1 and self.args.cache2) or (not self.args.cache1 and not self.args.cache2), \
            "Cache-to-cache comparison requires both --cache1 and --cache2;  otherwise neither for single cache comparison."

        if self.args.print_new_files:
            status = self.print_new_files()
        elif self.args.print_all_new_files:
            status = self.print_all_new_files()
        elif self.args.print_affected_instruments:
            status = self.print_affected_instruments()
        elif self.args.print_affected_types:
            status = self.print_affected_types()
        elif self.args.print_affected_modes:
            status = self.print_affected_modes()
        else:
            status = difference(self.observatory, self.old_file, self.new_file,
                                primitive_diffs=self.args.primitive_diffs,
                                check_diffs=self.args.check_diffs,
                                check_references=self.args.check_references,
                                mapping_text_diffs=self.args.mapping_text_diffs,
                                include_header_diffs=self.args.include_header_diffs,
                                hide_boring_diffs=self.args.hide_boring_diffs,
                                recurse_added_deleted=self.args.recurse_added_deleted,
                                lowest_mapping_only=self.args.lowest_mapping_only,
                                remove_paths=self.args.remove_paths,
                                squash_tuples=self.args.squash_tuples,
                                cache1=self.args.cache1,
                                cache2=self.args.cache2)
        if log.errors() or log.warnings():
            return 2
        else:
            return status

    def print_new_files(self):
        """Print the references or mappings which are in the second (new) context and not
        the firtst (old) context.
        """
        if not config.is_mapping(self.old_file) or not config.is_mapping(self.new_file):
            log.error("--print-new-files really only works for mapping differences.")
            return -1
        old = crds.get_pickled_mapping(self.old_file)   # reviewed
        new = crds.get_pickled_mapping(self.new_file)   # reviewed
        old_mappings = set(old.mapping_names())
        new_mappings = set(new.mapping_names())
        old_references = set(old.reference_names())
        new_references = set(new.reference_names())
        status = 0
        for name in sorted(new_mappings - old_mappings):
            print(name)
            status = 1
        for name in sorted(new_references - old_references):
            print(name)
            status = 1
        return status

    def print_all_new_files(self):
        """Print the names of all files which are in `new_file` (or any intermediary context) but not
        in `old_file`.   new_file > old_file.  Both new_file and old_file are similar mappings.
        """
        updated = get_updated_files(self.old_file, self.new_file)
        for mapping in updated:
            if config.is_mapping(mapping):
                print(mapping, self.instrument_filekind(mapping))
        for reference in updated:
            if not config.is_mapping(reference):
                print(reference, self.instrument_filekind(reference))
        return 1 if updated else 0

    def instrument_filekind(self, filename):
        """Return the instrument and filekind of `filename` as a space separated string."""
        instrument, filekind = utils.get_file_properties(self.observatory, filename)
        return instrument + " " + filekind

    def print_affected_instruments(self):
        """Print the instruments affected in a switch from `old_pmap` to `new_pmap`."""
        instrs = get_affected(self.old_file, self.new_file, include_header_diffs=self.args.include_header_diffs,
                              observatory=self.observatory)
        for instrument in sorted(instrs):
            print(instrument)
        return 1 if instrs else 0

    def print_affected_types(self):
        """Print the (instrument, filekind) pairs affected in a switch from `old_pmap` to `new_pmap`."""
        instrs = get_affected(self.old_file, self.new_file, include_header_diffs=self.args.include_header_diffs,
                              observatory=self.observatory)
        status = 0
        for instrument in sorted(instrs):
            for filekind in sorted(instrs[instrument]):
                if filekind.strip():
                    print("%-10s %-10s" % (instrument, filekind))
                    status = 1
        return status

    def print_affected_modes(self):
        """Print out all the affected mode tuples associated with the differences."""
        assert config.is_mapping(self.old_file) and config.is_mapping(self.new_file), \
            "for --print-affected-modes both files must be mappings."
        modes = mapping_affected_modes(self.old_file, self.new_file, self.args.include_header_diffs)
        for affected in modes:
            print(format_affected_mode(affected))
        return 1 if modes else 0

def test():
    """Run unit tests on this module."""
    import doctest
    from crds import diff
    return doctest.testmod(diff)

if __name__ == "__main__":
    sys.exit(DiffScript()())



"""This module codifies standard practices for scripted interactions with the
web server file submission system.
"""
import os
import io

from crds.core import log, utils
from crds.core.exceptions import CrdsError, CrdsWebError
from . import background

# from requests_toolbelt.multipart.encoder import MultipartEncoder, MultipartEncoderMonitor
DISABLED = []
try:
    import requests
except (ImportError, RuntimeError):
    log.verbose_warning("Import of 'requests' failed.  submit disabled.")
    DISABLED.append("requests")
try:
    from lxml import html
except (ImportError, RuntimeError):
    log.verbose_warning("Import of 'lxml' failed.  submit disabled.")
    DISABLED.append("lxml")

# ==================================================================================================

def log_section(section_name, section_value, verbosity=50, log_function=log.verbose,
                divider_name=None):
    """Issue log divider bar followed by a corresponding log message."""
    log.divider(name=divider_name, verbosity=verbosity, func=log.verbose)
    log_function(section_name, section_value, verbosity=verbosity+5)

# ==================================================================================================

_UPLOAD_CHUNK_SIZE = 2 * 1000 * 1000

class CrdsDjangoConnection:

    """This class handles CRDS authentication, basic GET, basic POST, and CRDS-style get/post.
    It also manages the CSRF token generated by Django to block form forgeries and CRDS instrument
    management/locking.
    """

    def __init__(self, locked_instrument="none", username=None, password=None, base_url=None):
        if DISABLED:
            raise CrdsError("Missing or broken depenencies:", DISABLED)
        self.locked_instrument = locked_instrument
        self.username = username
        self.password = password
        self.base_url = base_url
        self.session = requests.session()
        self.session.headers.update({'referer': self.base_url})

    def abs_url(self, relative_url):
        """Return the absolute server URL constructed from the given `relative_url`."""
        return self.base_url + relative_url

    def dump_response(self, name, response):
        """Print out verbose output related to web `response` from activity `name`."""
        log_section("headers:\n", response.headers, divider_name=name, verbosity=70)
        log_section("status_code:", response.status_code, verbosity=50)
        log_section("text:\n", response.text, verbosity=75)
        try:
            json_text = response.json()
            log_section("json:\n", json_text)
        except Exception:
            pass
        log.divider(func=log.verbose)

    def response_complete(self, args):
        """Wait for an aysnchronous web response, do debug logging,  check for errors."""
        response = background.background_complete(args)
        self.dump_response("Response: ", response)
        self.check_error(response)
        return response

    post_complete = get_complete = repost_complete = response_complete

    def get(self, relative_url):
        """HTTP(S) GET `relative_url` and return the requests response object."""
        args = self.get_start(relative_url)
        return self.get_complete(args)

    @background.background
    def get_start(self, relative_url):
        """Initiate a GET running in the background, do debug logging."""
        url = self.abs_url(relative_url)
        log_section("GET:", url, divider_name="GET: " + url.split("&")[0])
        return self.session.get(url)

    def post(self, relative_url, *post_dicts, **post_vars):
        """HTTP(S) POST `relative_url` and return the requests response object."""
        args = self.post_start(relative_url, *post_dicts, **post_vars)
        return self.post_complete(args)

    @background.background
    def post_start(self, relative_url, *post_dicts, **post_vars):
        """Initiate a POST running in the background, do debug logging."""
        url = self.abs_url(relative_url)
        vars = utils.combine_dicts(*post_dicts, **post_vars)
        log_section("POST:", vars, divider_name="POST: " + url)
        return self.session.post(url, data=vars)

    def repost(self, relative_url, *post_dicts, **post_vars):
        """First GET form from ``relative_url`,  next POST form to same
        url using composition of variables from *post_dicts and **post_vars.

        Maintain Django CSRF session token.
        """
        args = self.repost_start(relative_url, *post_dicts, **post_vars)
        return self.repost_complete(args)

    def repost_start(self, relative_url, *post_dicts, **post_vars):
        """Initiate a repost,  first getting the form synchronously and extracting
        the csrf token,  then doing a post_start() of the form and returning
        the resulting thread and queue.
        """
        response = self.get(relative_url)
        csrf_values= html.fromstring(response.text).xpath(
            '//input[@name="csrfmiddlewaretoken"]/@value'
            )
        if csrf_values:
            post_vars['csrfmiddlewaretoken'] = csrf_values[0]
        return self.post_start(relative_url, *post_dicts, **post_vars)

    """
    {'time_remaining': '3:57:58', 'user': 'jmiller_unpriv', 'created_on': '2017-02-23 16:12:55', 'type': 'instrument', 'is_expired': False, 'status': 'ok', 'name': 'miri'}
    """
    def fail_if_existing_lock(self):
        """Issue a warning if self.locked_instrument is already locked."""
        response = self.get("/lock_status/"+self.username+"/")
        log.verbose("lock_status:", response)
        json_dict = utils.Struct(response.json())
        if (json_dict.name and (not json_dict.is_expired) and (json_dict.type == "instrument") and (json_dict.user == self.username)):
            CrdsWebError("User", repr(self.username), "has already locked", repr(json_dict.name),
                         ".  Failing to avert collisions.  User --logout or logout on the website to bypass.")

    def login(self, next="/"):
        """Login to the CRDS website and proceed to relative url `next`."""
        self.session.cookies["ASB-AUTH"] = self.password
        response = self.repost(
            "/login/",
            username = self.username,
            password = self.password,
            instrument = self.locked_instrument,
            next = next,
            )
        self.check_login(response)

    def check_error(self, response):
        """Note an error + exception if response contains an error_message <div>."""
        self._check_error(response, '//div[@id="error_message"]', "CRDS server error:")
        self._check_error(response, '//div[@class="error_message"]', "CRDS server new form error:")

    def check_login(self, response):
        """Note an error + exception if response contains content indicating login error."""
        self._check_error(
            response, '//div[@id="error_login"]',
            "Error logging into CRDS server:")
        self._check_error(
            response, '//div[@id="error_message"]',
            "Error logging into CRDS server:")
        self._check_error(
            response, '//title[contains(text(), "MyST SSO Portal")]',
            "Error logging into CRDS server:")

    def _check_error(self, response, xpath_spec, error_prefix):
        """Extract the `xpath_spec` text from `response`,  if present issue a
        log ERROR with  `error_prefix` and the response `xpath_spec` text
        then raise an exception.  This may result in multiple ERROR messages.

        Issue a log ERROR for each form error,  then raise an exception
        if any errors found.

        returns None
        """
        errors = 0
        if response.ok:
            error_msg_parse = html.fromstring(response.text).xpath(xpath_spec)
            for parse in error_msg_parse:
                error_message = parse.text.strip().replace("\n","")
                if error_message:
                    if error_message.startswith("ERROR: "):
                        error_message = error_message[len("ERROR: ")]
                    errors += 1
                    log.error(error_prefix, error_message)
        else:
            log.error("CRDS server responded with HTTP error status", response.status_code)
            errors += 1

        if errors:
            raise CrdsWebError("A web transaction with the CRDS server had errors.")

    def logout(self):
        """Login to the CRDS website and proceed to relative url `next`."""
        self.get("/logout/")

    def upload_file(self, filepath):
        abs_url = self.abs_url("/upload/chunked/")
        response = self.session.get(abs_url)
        log.verbose("COOKIES:", log.PP(response.cookies))
        csrf_token = response.cookies["csrftoken"]
        file_size = os.stat(filepath).st_size
        filename = os.path.basename(filepath)

        if file_size < _UPLOAD_CHUNK_SIZE:
            files = {"files": (filename, open(filepath, "rb"))}
            data = {"csrfmiddlewaretoken": csrf_token}
            self.session.post(abs_url, files=files, data=data)
        else:
            with open(filepath, "rb") as f:
                start_byte = 0
                while True:
                    chunk = f.read(_UPLOAD_CHUNK_SIZE)
                    if len(chunk) == 0:
                        break

                    files = {"files": (filename, io.BytesIO(chunk))}
                    data = {"csrfmiddlewaretoken": csrf_token}
                    end_byte = start_byte + len(chunk) - 1
                    content_range = f"bytes {start_byte}-{end_byte}/{file_size}"
                    headers = {"Content-Range": content_range}
                    response = self.session.post(abs_url, files=files, data=data, headers=headers)
                    csrf_token = response.cookies["csrftoken"]
                    start_byte = end_byte + 1



"""
Tests for crds.core.crds_cache_locking which is nominally used to synchronize
access to the CRDS cache but is actually general purpose locking integrated
with
"""
import sys
import os
import time
import multiprocessing
import tempfile
import doctest

# ===================================================================

from crds.core import log, config, crds_cache_locking

# ===================================================================

from . import test_config

# ===================================================================

def multiprocessing_instance(output_file_name):
    """Pretend to do something generic."""
    output_file = open(output_file_name, "a")

    with crds_cache_locking.get_cache_lock():
        for char in "testing":
            output_file.write(char)
            output_file.flush()
            time.sleep(0.2)
        output_file.write("\n")
        output_file.flush()


def try_multiprocessing():
    """Run some test functions using multiprocessing."""
    # Starting with Python 3.8, the default start method in macOS is
    # "spawn", which causes the CRDS lock to be recreated for each
    # process.
    context = multiprocessing.get_context("fork")
    pool = context.Pool(5)
    with tempfile.NamedTemporaryFile(mode="a") as output_file:
        pool.map(multiprocessing_instance, [output_file.name]*5)
        pool.close()
        reader = open(output_file.name)
        print(reader.read())

def dt_default_locking():
    """
    Default locking configuration, enabled.

    >>> old_state = test_config.setup()
    >>> crds_cache_locking.init_locks()
    >>> _ = log.set_verbose()
    >>> crds_cache_locking.status()
    'enabled, multiprocessing'
    >>> try_multiprocessing()
    testing
    testing
    testing
    testing
    testing
    <BLANKLINE>
    >>> test_config.cleanup(old_state)
    """

def dt_multiprocessing_locking():
    """
    Default locking configuration, enabled.

    >>> old_state = test_config.setup()
    >>> _ = config.LOCKING_MODE.set("multiprocessing")
    >>> crds_cache_locking.init_locks()
    >>> _ = log.set_verbose()
    >>> crds_cache_locking.status()
    'enabled, multiprocessing'
    >>> try_multiprocessing()
    testing
    testing
    testing
    testing
    testing
    <BLANKLINE>
    >>> test_config.cleanup(old_state)
    """

def dt_filelock_locking():
    """
    Default locking configuration, enabled.

    >>> old_state = test_config.setup()
    >>> _ = config.LOCKING_MODE.set("filelock")
    >>> crds_cache_locking.init_locks()
    >>> _ = log.set_verbose()
    >>> crds_cache_locking.status()
    'enabled, filelock'
    >>> crds_cache_locking.get_cache_lock()
    CrdsFileLock('/tmp/crds.cache')
    >>> try_multiprocessing()
    testing
    testing
    testing
    testing
    testing
    <BLANKLINE>
    >>> test_config.cleanup(old_state)
    """

def dt_lockfile_locking():
    """
    Default locking configuration, enabled.

    >>> old_state = test_config.setup()
    >>> _ = config.LOCKING_MODE.set("lockfile")
    >>> crds_cache_locking.init_locks()
    >>> _ = log.set_verbose()
    >>> crds_cache_locking.status()
    'enabled, lockfile'
    >>> crds_cache_locking.get_cache_lock()
    CrdsLockFile('/tmp/crds.cache')

    >> try_multiprocessing()    XXXXX lockfile is broken
    testing
    testing
    testing
    testing
    testing
    <BLANKLINE>

    >>> test_config.cleanup(old_state)
    """

def dt_default_disabled():
    """
    Default locking configuration, disabled.   All bets are off
    for unlocked program output behavior...  just make sure it
    doesn't crash.

    >>> old_state = test_config.setup()
    >>> _ = config.USE_LOCKING.set(False)
    >>> _ = log.set_verbose()
    >>> crds_cache_locking.init_locks()
    CRDS - DEBUG -  CRDS_USE_LOCKING = False. Cannot support downloading CRDS files while multiprocessing.
    >>> crds_cache_locking.status()
    'disabled, multiprocessing'
    >>> doctest.ELLIPSIS_MARKER = '-ignore-'
    >>> try_multiprocessing() # doctest: +ELLIPSIS
    -ignore-
    >>> doctest.ELLIPSIS_MARKER = '...'
    >>> test_config.cleanup(old_state)
    """

def dt_default_readonly():
    """
    Default locking configuration, readonly cache, locking disabled.

    >>> old_state = test_config.setup()
    >>> _ = config.set_cache_readonly()
    >>> _ = log.set_verbose()
    >>> crds_cache_locking.init_locks()
    >>> crds_cache_locking.status()
    'disabled, multiprocessing'
    >>> doctest.ELLIPSIS_MARKER = '-ignore-'
    >>> try_multiprocessing() # doctest: +ELLIPSIS
    -ignore-
    >>> doctest.ELLIPSIS_MARKER = '...'
    >>> test_config.cleanup(old_state)
    """

# ==================================================================================

def main():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_locking, tstmod
    return tstmod(test_locking)

if __name__ == "__main__":
    print(main())



"""This module is a command line script which dowloads the references and
mappings required to support a set of contexts from the CRDS server:

Old references and mappings which are no longer needed can be automatically
removed by specifying --purge-mappings or --purge-references::

  % crds sync --range 1:2 --purge-mappings --purge-references

will remove references or mappings not required by hst_0001.pmap or
hst_0002.pmap in addition to downloading the required files.

Or explicitly list the files you want cached:

  % crds sync --files <references or mappings to cache>

To sync best references and rules for specific dataset FITS files:

  % crds sync --contexts hst_0001.pmap hst_0002.pmap --dataset-files *.fits --fetch-references

To sync best references and rules for specific dataset ids:

  % crds sync --contexts hst_0001.pmap hst_0002.pmap --dataset-ids J6M915030 --fetch-references

"""
import sys
import os
import os.path
import re
import shutil
import glob

# ============================================================================

import crds
from crds.core import log, config, utils, rmap, heavy_client, cmdline, crds_cache_locking
from crds import data_file
from crds.core.log import srepr
from crds.client import api

# ============================================================================

# ============================================================================

class SyncScript(cmdline.ContextsScript):
    """Command line script for synchronizing local CRDS file cache with CRDS server."""

    description = """
    Synchronize local mapping and reference caches for the given contexts by
    downloading missing files from the CRDS server and/or archive.
    """

    epilog = """
    * Dry-Running Cache Changes

       Since CRDS cache operations can involve significant network downloads,  as a general note,
       crds.sync can be run with *---readonly-cache ---verbose* switches to better determine what
       the effects of any command should be.   This can be used to gauge download sizes or list
       files before deleting them.

    * Syncing Specific Files

        Downloading an explicit list of files can be done by like this::

        % crds sync  --files hst_0001.pmap hst_acs_darkfile_0037.fits

        this will download only those two files into the appropriate locations
        in your CRDS cache.

        An output directory outside the CRDS cache can also be specified like
        this::

        % crds sync --output-dir . --files hst_0001.pmap hst_acs_darkfile_0037.fits

        which will fetch the same files but put them in the current working
        directory "." instead of at their implicit locations in the CRDS cache.

    * Syncing Rules

        Typically syncing CRDS files is done with respect to particular CRDS contexts:

        Synced contexts can be explicitly listed::

            % crds sync  --contexts hst_0001.pmap hst_0002.pmap

        this will recursively download all the mappings referred to by .pmaps 0001 and 0002.

        Synced contexts can be specified as a numerical range::

            % crds sync --range 1:3

        this will also recursively download all the mappings referred to by .pmaps 0001, 002, 0003.

        Synced contexts can be specified as --all contexts::

            % crds sync --all

        this will recursively download all CRDS mappings for all time.

    * Syncing References By Context

        Because complete reference downloads can be enormous,  you must explicitly specify when
        you wish to fetch the references which are enumerated in particular CRDS rules::

            % crds sync  --contexts hst_0001.pmap hst_0002.pmap  --fetch-references

        will download all the references mentioned by contexts 0001 and 0002.

        This can be a huge (1T+) network download and should generally only be
        used by institutions,  not individual researchers.

        **NOTE:** the contexts synced can be for particular instruments or types rather than
        the entire pipeline,  e.g. hst_cos_0002.imap or hst_cos_proftab_0001.rmap

    * Removing Unused Files

        CRDS rules from **unspecified** contexts can be removed like this::

            % crds sync  --contexts hst_0004.pmap hst_0005.pmap --purge-mappings

        while this would remove references which are *not* in contexts 4 or 5::

            % crds sync  --contexts hst_0004.pmap hst_0005.pmap --purge-references

        Again, both of these commands remove cached files which are not specified or implied.

    * References for Dataset Files

        References required by particular dataset files can be cached like this::

            % crds sync  --contexts hst_0001.pmap hst_0002.pmap --dataset-files  <dataset_files...> e.g. acs_J8D219010.fits

        This will fetch all the references required to support the listed datasets for contexts 0001 and 0002.

        This mode does not update dataset file headers.  See also crds.bestrefs for similar functionality with header updates.

    * References for Dataset Ids

        References for particular dataset ids can be cached like this::

            % crds sync  --contexts hst_0001.pmap hst_0002.pmap --dataset-ids  <ids...>  e.g. J6M915030

        This will fetch all the references required to support the listed dataset ids for contexts 0001 and 0002.

    * Checking and Repairing Large Caches

        Large Institutional caches can be checked and/or repaired like this::

            % crds sync --contexts hst_0001.pmap --fetch-references --check-sha1sum --repair-files

        will download all the files in hst_0001.pmap not already present.

        Both mappings and references would then be checked for correct length, sha1sum, and status.

        Any files with bad length or checksum would then be deleted and re-downloaded.   This is really intended
        for a large *existing* cache.

        File checksum verification is optional because it is time consuming.  Verifying the contents of the current
        HST shared cache requires 8-10 hours.   In contrast, doing simple length, existence, and status checks
        takes 5-10 minutes,  sufficient for a quick check but not foolproof.

    * Checking Smaller Caches,  Identifying Foreign Files

        The simplest approach for "repairing" a small cache is to delete it and resync::

           % rm -rf $CRDS_PATH
           % crds sync  -- ...  # repeat whatever syncs you did to cache files of interest

        A more complicated but also more precise approach to check all rules and references in your local CRDS cache::

           % crds sync --check-files --files `crds list --all --cached-mappings --cached-references`

        This approach works by using the crds.list command to dump the file names of all files in the CRDS cache
        and then using the crds.sync command to check exactly those files.

        Since --cached-mappings and --cached-references will only print the name of rules or references in the cache,
        not all files from CRDS,  the second approach can also be used to detect files which are not from CRDS.

        For smaller caches *--check-sha1sum* is likekly to be less of a performance/runtime issue and should be used
        to detect files which have changed in contents but not in length,  particularly CRDS mapping files.

    * Removing blacklisted or rejected files

        crds.sync can be used to remove the files from specific contexts which have been marked as "bad".

          % crds sync --contexts hst_0001.pmap --fetch-references --check-files --purge-rejected --purge-blacklisted

        would first sync the cache downloading all the files in hst_0001.pmap.  Both mappings and references would then
        be checked for correct length.   Files reported as rejected or blacklisted by the server would be removed.

    * Reorganizing cache structure

        CRDS now supports two cache structures for organizing references: flat and instrument.  *flat* places all references
        for a telescope in a single directory,  e.g. references/hst.   *instrument* segregates references into subdirectories
        which name instruments or legacy environment variables,  e.g. acs or jref.

        Newly created caches will default to the *instrument* organization.  To migrate a legacy cache with a flat single
        directory layout to the new structure,  sync with --organize=instrument::

           % crds sync --organize=instrument --verbose

        To migrate to the flat structure,  use --organize=flat::

           % crds sync --organize=flat --verbose

        While reorganizing, if CRDS makes note of "junk files" in your cache which are
        obstructing the process of reorganizing, you can allow CRDS to delete the junk
        by adding --organize-delete-junk.

        The --organize switches are intended to be used only on inactive file caches
        when calibration software is not running and actively using CRDS.
    """

    # ------------------------------------------------------------------------------------------

    def add_args(self):
        super(SyncScript, self).add_args()
        self.add_argument("--files", nargs="*", help="Explicitly list files to be synced.")
        self.add_argument('--dataset-files', metavar='DATASET', type=cmdline.dataset, nargs='*',
                          help='Cache references for the specified datasets FITS files.')
        self.add_argument('--dataset-ids', metavar='DATASET', type=str, nargs='*',
                          help='Cache references for the specified dataset ids.')
        self.add_argument('--fetch-references', action='store_true', dest="fetch_references",
                          help='Cache all the references for the specified contexts.')
        self.add_argument('--purge-references', action='store_true', dest="purge_references",
                          help='Remove reference files not referred to by contexts from the cache.')
        self.add_argument('--purge-mappings', action='store_true', dest="purge_mappings",
                          help='Remove mapping files not referred to by contexts from the cache.')
        self.add_argument('--dry-run', action="store_true",
                          help= "Don't remove purged files, or repair files,  just print out their names.")

        self.add_argument('-k', '--check-files', action='store_true', dest='check_files',
                          help='Check cached files against the CRDS database and report anomalies.')
        self.add_argument('-s', '--check-sha1sum', action='store_true', dest='check_sha1sum',
                          help='For --check-files,  also verify file sha1sums.')
        self.add_argument('-r', '--repair-files', action='store_true', dest='repair_files',
                          help='Repair or re-download files noted as bad by --check-files')
        self.add_argument('--purge-rejected', action='store_true', dest='purge_rejected',
                          help='Purge files noted as rejected by --check-files')
        self.add_argument('--purge-blacklisted', action='store_true', dest='purge_blacklisted',
                          help='Purge files (and their mapping anscestors) noted as blacklisted by --check-files')
        self.add_argument('--fetch-sqlite-db', action='store_true', dest='fetch_sqlite_db',
                          help='Download a sqlite3 version of the CRDS file catalog.')
        self.add_argument("--organize", metavar="NEW_SUBDIR_MODE", type=config.check_crds_ref_subdir_mode,
                          nargs="?", default=None,
                          help="Migrate cache to specified structure, 'flat' or 'instrument'. WARNING: perform only on idle caches.")
        self.add_argument("--organize-delete-junk", action="store_true",
                          help="When --organize'ing, delete obstructing files or directories CRDS discovers.")
        self.add_argument("--verify-context-change", action="store_true",
                          help="Make it an error if the context does not update to something new.")
        self.add_argument("--push-context", metavar="KEY", type=str,
                          help="Push the name of the final cached context to the server for the pipeline identified by KEY.")
        self.add_argument("--clear-pickles", action="store_true",
                          help="Remove all context pickles from the CRDS cache. Can precede --save-pickles.")
        self.add_argument("--save-pickles", action="store_true",
                          help="Save pre-compiled versions of the sync'ed contexts in the CRDS cache.  Keep pre-existing pickles.")
        self.add_argument("--output-dir", type=str, default=None,
                          help="Directory to output sync'ed files, for simple syncs,  particularly --files.   Implies 'flat' cache.")
        self.add_argument("--clear-locks", action="store_true",
                          help="Remove CRDS cache file lock(s).")
        self.add_argument("--force-config-update", action="store_true",
                          help="Even if sync errors occur, attempt to update the CRDS configuration, including the default context.")

    # ------------------------------------------------------------------------------------------

    def main(self):
        """Synchronize files."""

        # clear any mutliprocessing locks associated with the CRDS cache.
        if self.args.clear_locks:
            crds_cache_locking.clear_cache_locks()
            return log.errors()

        self.handle_misc_switches()   # simple side effects only

        # if explicitly requested,  or the cache is suspect or being ignored,  clear
        # cached context pickles.
        if self.args.clear_pickles or self.args.ignore_cache or self.args.repair_files:
            self.clear_pickles()

        # utility to change cache structure, e.g. add instrument subdirs.
        # do this before syncing anything under the current mode.
        if self.args.organize:
            self.organize_references(self.args.organize)

        # fetching and verifying files both require a server connection.
        self.require_server_connection()

        # primary sync'ing occurs here,  both mappings and references as well as odd-ball
        # server sqlite3 database download.
        verify_file_list = self.file_transfers()

        # verification is relative to sync'ed files,  and can include file replacement if
        # defects are found.
        if (self.args.check_files or self.args.check_sha1sum or self.args.repair_files or
            self.args.purge_blacklisted or self.args.purge_rejected):
            self.verify_files(verify_file_list)

        # context pickles should only be (re)generated after mappings are fully sync'ed and verified
        if self.args.save_pickles:
            self.pickle_contexts(self.contexts)

        # update CRDS cache config area,  including stored version of operational context.
        # implement pipeline support functions of context update verify and echo
        # If --output-dir was specified,  do not update cache config.
        if self.args.output_dir:
            log.verbose_warning("Used --output-dir,  skipping cache server_config update including default context and bad files.")
            if config.writable_cache_or_verbose("skipping removal of ref_cache_subdir_mode file."):
                os.remove(config.get_crds_ref_subdir_file(self.observatory))
        else:
            self.update_context()

        self.report_stats()
        log.standard_status()
        return log.errors()
    # ------------------------------------------------------------------------------------------

    def handle_misc_switches(self):
        """Handle command line switches with simple side-effects that should precede
        other sync operations.
        """
        if self.args.dry_run:
            config.set_cache_readonly(True)

        if self.args.repair_files:
            self.args.check_files = True

        if self.args.output_dir:
            os.environ["CRDS_MAPPATH_SINGLE"] = self.args.output_dir
            os.environ["CRDS_REFPATH_SINGLE"] = self.args.output_dir
            os.environ["CRDS_CFGPATH_SINGLE"] = self.args.output_dir
            os.environ["CRDS_PICKLEPATH_SINGLE"] = self.args.output_dir
            self.args.organize = "flat"

        if self.readonly_cache:
            log.info("Syncing READONLY cache,  only checking functions are enabled.")
            log.info("All cached updates, context changes, and file downloads are inhibited.")

    def file_transfers(self):
        """Top level control for the primary function of downloading files specified as:

        --files ...      (explicit list of CRDS mappings or references)
        --contexts ...   (many varieties of mapping specifier including --all, --range, etc.)
        --fetch-sqlite-db ...  (Server catalog download as sqlite3 database file.

        Returns list of downloaded/cached files for later verification if requested.
        """
        if self.args.files:
            self.sync_explicit_files()
            verify_file_list = self.files
        elif self.args.fetch_sqlite_db:
            self.fetch_sqlite_db()
            verify_file_list = []
        elif self.contexts:
            verify_file_list = self.interpret_contexts()
        else:
            log.error("Define --all, --contexts, --last, --range, --files, or --fetch-sqlite-db to sync.")
            sys.exit(-1)
        return verify_file_list

    def interpret_contexts(self):
        """Sync the mapping files associated with any context specifiers like
        --all, --contexts,  etc...

        If --fetch-references or --purge-references are specified, also fetch and/or
        purge references with respect to the specified contexts.
        """
        active_mappings = self.get_context_mappings()
        verify_file_list = active_mappings
        if self.args.fetch_references or self.args.purge_references:
            active_references = self.get_synced_references()
            if self.args.purge_rejected or self.args.purge_blacklisted:
                references = active_references - self.bad_files
            else:
                references = active_references
            if self.args.fetch_references:
                self.fetch_files(self.contexts[0], references)
                verify_file_list += active_references
            if self.args.purge_references:
                self.purge_references(active_references)
        if self.args.purge_mappings:
            self.purge_mappings()
        return verify_file_list

    def get_synced_references(self):
        """Return the list of reference names associated with the specified dataset
        files, dataset ids, or contexts, including any associated GEIS data
        file names.
        """
        if self.args.dataset_files or self.args.dataset_ids:
            active_references = self.sync_datasets()
        else:
            active_references = self.get_context_references()
        # Handle GEIS paired data files
        active_references = set(active_references + self.get_conjugates(active_references))
        return active_references

    def update_context(self):
        """Update the CRDS operational context in the cache.  Handle pipeline-specific
        targeted features of (a) verifying a context switch as actually recorded in
        the local CRDS cache and (b) echoing/pushing the pipeline context back up to the
        CRDS server for tracking using an id/authorization key.

        If errors occurred during the sync and --force_config_update is not set,
        """
        if not log.errors() or self.args.force_config_update:
            if self.args.verify_context_change:
                old_context = heavy_client.load_server_info(self.observatory).operational_context
            heavy_client.update_config_info(self.observatory)
            if self.args.verify_context_change:
                self.verify_context_change(old_context)
            if self.args.push_context:
                self.push_context()
        else:
            log.warning("Errors occurred during sync,  skipping CRDS cache config and context update.")

    def clear_pickles(self):
        """Remove all pickles."""
        log.info("Removing all context pickles.  Use --save-pickles to recreate for specified contexts.")
        for path in rmap.list_pickles("*.pmap", self.observatory, full_path=True):
            if os.path.exists(path):
                utils.remove(path, self.observatory)

    def pickle_contexts(self, contexts):
        """Save pickled versions of `contexts` in the CRDS cache.

        By default this will by-pass existing pickles if they successfully load.
        """
        for context in contexts:
            with log.error_on_exception("Failed pickling", repr(context)):
                crds.get_pickled_mapping.uncached(context, use_pickles=True, save_pickles=True)  # reviewed

    # ------------------------------------------------------------------------------------------

    @property
    def server_info(self):
        """Return the server_info dict from the CRDS server.  Do not call update_config_info() until sync complete."""
        return heavy_client.get_config_info(self.observatory)

    def verify_context_change(self, old_context):
        """Verify that the starting and post-sync contexts are different,  or issue an error."""
        new_context = heavy_client.load_server_info(self.observatory).operational_context
        if old_context == new_context:
            log.error("Expected operational context switch but starting and post-sync contexts are both", repr(old_context))
        else:
            log.info("Operational context updated from", repr(old_context), "to",  repr(new_context))

    def push_context(self):
        """Push the final context recorded in the local cache to the CRDS server so it can be displayed
        as the operational state of a pipeline.
        """
        info = heavy_client.load_server_info(self.observatory)
        with log.error_on_exception("Failed pushing cached operational context name to CRDS server"):
            api.push_remote_context(self.observatory, "operational", self.args.push_context, info.operational_context)
            log.info("Pushed cached operational context name", repr(info.operational_context), "to CRDS server")

    # ------------------------------------------------------------------------------------------

    def purge_mappings(self):
        """Remove all mappings not under pmaps `self.contexts`."""
        # rmap.list_mappings lists all mappings in the *local* cache.
        # in contrast,  client.list_mappings globs available mappings in the server cache.
        purge_maps = set(rmap.list_mappings('*.[pir]map', self.observatory))
        keep = set(self.get_context_mappings())
        self.remove_files(sorted(purge_maps-keep), "mapping")

    # ------------------------------------------------------------------------------------------

    def fetch_files(self, context, files):
        """Downloads `files` as needed relative to `context` nominally used to identify
        the observatory to the server.  If the CRDS cache is currently configured as READONLY,
        prints an estimate about which files will download and the total size.  The estimate
        does not (currently) include the nested mappings associated with the closure of `files`,
        but the dominant costs of downloads are reference files.
        """
        files = set([os.path.basename(_file) for _file in files])
        if config.get_cache_readonly():
            log.info("READONLY CACHE estimating required downloads.")
            if not self.args.ignore_cache:
                already_have = (set(rmap.list_references("*", self.observatory)) |
                                set(rmap.list_mappings("*", self.observatory)))
            else:
                already_have = set()
            fetched = [ x for x in sorted(files - already_have) if not x.startswith("NOT FOUND") ]
            for _file in files:
                if _file in already_have:
                    log.verbose("File", repr(_file), "is already in the CRDS cache.", verbosity=55)
                else:
                    log.verbose("File", repr(_file), "would be downloaded.", verbosity=55)
            if fetched:
                with log.info_on_exception("File size information not available."):
                    info_map = api.get_file_info_map(self.observatory, fetched, fields=["size"])
                    total_bytes = api.get_total_bytes(info_map)
                    log.info("READONLY CACHE would download", len(fetched), "files totalling",
                             utils.human_format_number(total_bytes).strip(), "bytes.")
            else:
                log.info("READONLY CACHE no reference downloads expected.")
        else:
            self.dump_files(context, files, self.args.ignore_cache)

    def purge_references(self, keep=None):
        """Remove all references not references under pmaps `self.contexts`."""
        purge_refs = set(rmap.list_references("*", self.observatory))
        if keep is None:
            keep = set(self.get_context_references())
        else:
            keep = set(keep)
        self.remove_files(sorted(purge_refs - keep), "reference")

    def remove_files(self, files, kind):
        """Remove the list of `files` basenames which are converted to fully
        specified CRDS paths using the locator module associated with context.
        """
        if not files:
            log.verbose("No " + kind + "s to remove.")
        files2 = set(files)
        for filename in files:
            if re.match(r"\w+\.r[0-9]h", filename):
                files2.add(filename[:-1] + "d")
        for filename in files:
            with log.error_on_exception("Failed purging", kind, repr(filename)):
                where = config.locate_file(filename, self.observatory)
                utils.remove(where, observatory=self.observatory)

    # ------------------------------------------------------------------------------------------

    def sync_datasets(self):
        """Sync mappings and references for datasets with respect to `self.contexts`."""
        if not self.contexts:
            log.error("Define --contexts under which references are fetched for --dataset-files or --dataset-ids.""")
            sys.exit(-1)
        active_references = []
        for context in self.contexts:
            if self.args.dataset_ids:
                if len(self.args.dataset_ids) == 1 and self.args.dataset_ids[0].startswith("@"):
                    with open(self.args.dataset_ids[0][1:]) as pfile:
                        self.args.dataset_ids = pfile.read().splitlines()
                with log.error_on_exception("Failed to get matching parameters for", self.args.dataset_ids):
                    id_headers = api.get_dataset_headers_by_id(context, self.args.dataset_ids)
            for dataset in self.args.dataset_files or self.args.dataset_ids:
                log.info("Syncing context '%s' dataset '%s'." % (context, dataset))
                with log.error_on_exception("Failed to get matching parameters from", repr(dataset)):
                    if self.args.dataset_files:
                        headers = { dataset : data_file.get_conditioned_header(dataset, observatory=self.observatory) }
                    else:
                        headers = { dataset_id : header for (dataset_id, header) in id_headers.items() if
                                    dataset.upper() in dataset_id }
                    for assc_dataset, header in headers.items():
                        with log.error_on_exception("Failed syncing references for dataset", repr(assc_dataset),
                                                    "under context", repr(context)):
                            bestrefs = crds.getrecommendations(header, context=context, observatory=self.observatory,
                                                               ignore_cache=self.args.ignore_cache)
                            log.verbose("Best references for", repr(assc_dataset), "are", bestrefs)
                            active_references.extend(bestrefs.values())
        active_references = [ ref for ref in active_references if not ref.startswith("NOT FOUND") ]
        log.verbose("Syncing references:", repr(active_references))
        return list(set(active_references))

    # ------------------------------------------------------------------------------------------

    def sync_explicit_files(self):
        """Cache `self.files`."""
        log.info("Syncing explicitly listed files.")
        self.fetch_files(self.default_context, self.files)

    # ------------------------------------------------------------------------------------------

    def verify_files(self, files):
        """Check `files` against the CRDS server database to ensure integrity and check reject status."""
        basenames = [os.path.basename(file) for file in files]
        try:
            log.verbose("Downloading verification info for", len(basenames), "files.", verbosity=10)
            infos = api.get_file_info_map(observatory=self.observatory, files=basenames,
                                         fields=["size","rejected","blacklisted","state","sha1sum"])
        except Exception as exc:
            log.error("Failed getting file info.  CACHE VERIFICATION FAILED.  Exception: ", repr(str(exc)))
            return
        bytes_so_far = 0
        total_bytes = api.get_total_bytes(infos)
        for nth_file, file in enumerate(files):
            bfile = os.path.basename(file)
            if infos[bfile] == "NOT FOUND":
                log.error("CRDS has no record of file", repr(bfile))
            else:
                self.verify_file(file, infos[bfile], bytes_so_far, total_bytes, nth_file, len(files))
                bytes_so_far += int(infos[bfile]["size"])

    def verify_file(self, file, info, bytes_so_far, total_bytes, nth_file, total_files):
        """Check one `file` against the provided CRDS database `info` dictionary."""
        path = config.locate_file(file, observatory=self.observatory)
        base = os.path.basename(file)
        n_bytes = int(info["size"])

        # Only output verification info for slow sha1sum checks by default
        log.verbose(
            api.file_progress(
                "Verifying", base, path, n_bytes, bytes_so_far, total_bytes, nth_file, total_files),
            verbosity=10 if self.args.check_sha1sum else 60)

        if not os.path.exists(path):
            if base not in self.bad_files:
                log.error("File", repr(base), "doesn't exist at", repr(path))
            return

        # Checks which force repairs should do if/else to avoid repeat repair
        size = os.stat(path).st_size
        if int(info["size"]) != size:
            self.error_and_repair(path, "File", repr(base), "length mismatch LOCAL size=" + srepr(size),
                                  "CRDS size=" + srepr(info["size"]))
        elif self.args.check_sha1sum or config.is_mapping(base):
            log.verbose("Computing checksum for", repr(base), "of size", repr(size), verbosity=60)
            sha1sum = utils.checksum(path)
            if info["sha1sum"] == "none":
                log.warning("CRDS doesn't know the checksum for", repr(base))
            elif info["sha1sum"] != sha1sum:
                self.error_and_repair(path, "File", repr(base), "checksum mismatch CRDS=" + repr(info["sha1sum"]),
                                      "LOCAL=" + repr(sha1sum))

        if info["state"] not in ["archived", "operational"]:
            log.warning("File", repr(base), "has an unusual CRDS file state", repr(info["state"]))

        if info["rejected"] != "false":
            log.verbose_warning("File", repr(base), "has been explicitly rejected.", verbosity=60)
            if self.args.purge_rejected:
                self.remove_files([path], "file")
            return

        if info["blacklisted"] != "false":
            log.verbose_warning("File", repr(base), "has been blacklisted or is dependent on a blacklisted file.",
                                verbosity=60)
            if self.args.purge_blacklisted:
                self.remove_files([path], "file")
            return
        return

    def error_and_repair(self, file, *args, **keys):
        """Issue an error message and repair `file` if requested by command line args."""
        log.error(*args, **keys)
        if self.args.repair_files:
            if config.writable_cache_or_info("Skipping remove and re-download of", repr(file)):
                log.info("Repairing file", repr(file))
                utils.remove(file, observatory=self.observatory)
                self.dump_files(self.default_context, [file])

    def fetch_sqlite_db(self):
        """Download a SQLite version of the CRDS catalog from the server."""
        path = api.get_sqlite_db(self.observatory)
        log.info("SQLite database file downloaded to:", path)

    def organize_references(self, new_mode):
        """Find all references in the CRDS cache and relink them to the paths which are implied by `new_mode`.
        This is used to reroganize existing file caches into new layouts,  e.g. flat -->  by instrument.
        """
        old_refpaths = rmap.list_references("*", observatory=self.observatory, full_path=True)
        old_mode = config.get_crds_ref_subdir_mode(self.observatory)
        log.info("Reorganizing", len(old_refpaths), "references from", repr(old_mode), "to", repr(new_mode))
        config.set_crds_ref_subdir_mode(new_mode, observatory=self.observatory)
        new_mode = config.get_crds_ref_subdir_mode(self.observatory)  # did it really change.
        for refpath in old_refpaths:
            with log.error_on_exception("Failed relocating:", repr(refpath)):
                desired_loc = config.locate_file(os.path.basename(refpath), observatory=self.observatory)
                if desired_loc != refpath:
                    if os.path.exists(desired_loc):
                        if not self.args.organize_delete_junk:
                            log.warning("Link or directory already exists at", repr(desired_loc), "Skipping", repr(refpath))
                            continue
                        utils.remove(desired_loc, observatory=self.observatory)
                    if config.writable_cache_or_info("Skipping file relocation from", repr(refpath), "to", repr(desired_loc)):
                        log.info("Relocating", repr(refpath), "to", repr(desired_loc))
                        shutil.move(refpath, desired_loc)
                else:
                    if old_mode != new_mode:
                        log.verbose_warning("Keeping existing cached file", repr(desired_loc), "already in target mode", repr(new_mode))
                    else:
                        log.verbose_warning("No change in subdirectory mode", repr(old_mode), "skipping reorganization of", repr(refpath))
        if new_mode == "flat" and old_mode == "instrument":
            log.info("Reorganizing from 'instrument' to 'flat' cache,  removing instrument directories.")
            for instrument in self.locator.INSTRUMENTS:
                self.remove_dir(instrument)

    def remove_dir(self, instrument):
        """Remove an instrument cache directory and any associated legacy link."""
        if config.writable_cache_or_info("Skipping remove instrument", repr(instrument), "directory."):
            crds_refpath = config.get_crds_refpath(self.observatory)
            prefix = self.locator.get_env_prefix(instrument)
            rootdir = os.path.join(crds_refpath, instrument)
            refdir = os.path.join(crds_refpath, prefix[:-1])
            if len(glob.glob(os.path.join(rootdir, "*"))):
                log.info("Residual files in '{}'. Not removing.".format(rootdir))
                return
            if os.path.exists(refdir):   # skip crds://  vs.  oref
                utils.remove(refdir, observatory=self.observatory)
            utils.remove(rootdir, observatory=self.observatory)

# ==============================================================================================================

if __name__ == "__main__":
    sys.exit(SyncScript()())



"""
The crds.client package is a light weight web client which accesses
the CRDS server using jsonrpc to compute best references and obtain reference
files.   Although it has some dependencies on the core crds package,  the client
module tries to minimize them to lightweight functions from: config, utils,
and log.
"""
from .api import *



import sys
import os.path

from crds.core import log, utils, config, timestamp, cmdline
from crds.client import api

class QueryAffectedDatasetsScript(cmdline.Script):

    description = """
query_affected_datasets (QAD) queries the CRDS server for datasets affected by the specified
CONTEXT(s), history INDICES, or DATES.  QAD relies on pre-computed results on the server,
so QAD queries are only valid for historical context transitions from the context history.
When no history interval is specified,  QAD uses the last processed context as the starting
point,  and the end of the current history as the stopping point.

Before going too far, a couple of points are in order:

    1. QueryAffectedDatasetsScript is intended to be sub-classed.   By default it just prints dataset
    IDs on STDOUT as well as log and affected datasets computation output to STDERR.   Override the
    use_affected() and use_all_ids() methods to do something custom with the affected dataset IDs,
    either switch-by-switch, or potentially multiple contexts at once, respectively.   Override the
    log_affected() or log_all_ids() methods to customize most logging.   QAD provides basic interaction
    with the context history, affected datasets service, recording state, and error handling.

    2. crds.bestrefs can be used to compute the datasets affected by arbitrary context transitions
    when run locally.   The computation can require minutes to hours depending on number of instruments,
    types, datasets, and tables potentially affected.    For this, the --affected-datasets switch
    configures crds.bestrefs to perform an affected datasets computation using the bundle of standard
    options run on the CRDS servers.

Due to processing order on the server, new contexts appear in the context history as operational
before the datasets affected have been computed.   This framework helps resolve that race condition
and provides options for handling affected datasets computations which contained errors.

To support interactive experimentation, QAD supports listing the context history:

    % query_affected_datasets --list
    (0, '2013-07-02 15:44:53', 'hst.pmap', 'set by system')
    (1, '2013-09-10 18:23:06', 'hst_0003.pmap', 'Updated hst.pmap with new references (known to reffile_ops_rep on harpo) up to 09-10-2013')
    ...
    (85, '2014-09-23 16:48:17', 'hst_0287.pmap', 'Delivery of new WFC3 darks.')
    (86, '2014-10-13 11:26:40', 'hst_0288.pmap', 'Delivery of a new ACS WFC1-1K bias.')
    (87, '2014-10-14 09:34:29', 'hst_0289.pmap', 'Delivery of a new COS FUV BPIXTAB.')

See also the -x and -y parameters below for customizing interactive query ranges.

With no history range specified, QAD selects the last history item processed as the starting point
and the end of the current history as the stopping point.  Normally there's nothing new to report,
the last thing processed was the end of the history and nothing has changed on the server.

    % query_affected_datasets
    CRDS - INFO - No new results available.
    CRDS - INFO - 0 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 1 infos

Following a context change,  by default QAD will notice a difference between the last saved context
and the new last context in the history.  QAD is designed to be sub-classed but by default prints log
information and recorded affected datasets output to STDERR.   It prints affected dataset IDs to STDOUT.
Run periodically,  QAD will typically see at most a single context switch.

    % query_affected_datasets > ids
    CRDS - INFO - Fetching effects for (96, '2014-11-25 15:48:40', 'hst_0300.pmap', 'Delivery of a new COS HVTAB for association LCIX02080.')
    ####################################################################################################
    --------------------------------------------------------------------------------------------------------------
    CRDS hst ops datasets affected hst_0299.pmap --> hst_0300.pmap on 2014-11-25-15:50:09
    --------------------------------------------------------------------------------------------------------------
    CRDS - INFO - [2014-11-25 15:50:12,563]  Mapping differences from 'hst_0299.pmap' --> 'hst_0300.pmap' affect:
     {'cos': ['hvtab']}
    CRDS - INFO - [2014-11-25 15:50:12,731]  Possibly affected --datasets-since dates determined by 'hst_0299.pmap' --> 'hst_0300.pmap' are:
     {'cos': '2009-05-11 00:00:00'}
    CRDS - INFO - [2014-11-25 15:50:12,731]  Computing bestrefs for db datasets for ['cos']
    CRDS - INFO - [2014-11-25 15:50:12,731]  Dumping dataset parameters for 'cos' from CRDS server at 'https://hst-crds.stsci.edu' since '2009-05-11 00:00:00'
    CRDS - INFO - [2014-11-25 15:50:16,457]  Downloaded  19592 dataset ids for 'cos' since '2009-05-11 00:00:00'
    CRDS - INFO - [2014-11-25 15:51:16,293]  Updated exposure counts:
     {'COS': {'hvtab': 13696}}
    CRDS - INFO - [2014-11-25 15:51:16,309]  Affected products = 9494
    CRDS - INFO - [2014-11-25 15:51:16,319]  Unique error types: 0
    CRDS - INFO - [2014-11-25 15:51:16,319]  STARTED 2014-11-25 15:50:10.67
    CRDS - INFO - [2014-11-25 15:51:16,319]  STOPPED 2014-11-25 15:51:16.31
    CRDS - INFO - [2014-11-25 15:51:16,319]  ELAPSED 0:01:05.64
    CRDS - INFO - [2014-11-25 15:51:16,319]  19.59 K datasets at 298.47  datasets-per-second
    CRDS - INFO - [2014-11-25 15:51:16,319]  0 errors
    CRDS - INFO - [2014-11-25 15:51:16,319]  0 warnings
    CRDS - INFO - [2014-11-25 15:51:16,319]  12 infos
    --------------------------------------------------------------------------------------------------------------
    OK: CRDS hst ops datasets affected hst_0299.pmap --> hst_0300.pmap on 2014-11-25-15:50:09 : 9494 affected
    --------------------------------------------------------------------------------------------------------------
    ####################################################################################################
    CRDS - INFO - Contributing context switches = 1
    CRDS - INFO - Total products affected = 9494
    CRDS - INFO - 0 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 3 infos

    % cat ids
    i9zf01010
    i9zf02010
    i9zf03010
    i9zf04010
    i9zf05010
    ...

STDERR output from multiple context switches is delimited by ########## lines.

For the sake of custom queries,  QAD supports --starting-context (-x) and --stopping-context (-y) parameters to define the
history starting and stopping points.  -x and -y can be specified as CONTEXTS, HISTORY INDICES, or DATES.

    % query_affected_datasets -x 92 -y 94 > ids
    CRDS - INFO - Fetching effects for (92, '2014-11-05 13:53:15', 'hst_0296.pmap', 'Delivery of new ACS bias, dark, and cte_corrected dark reference files.')
    CRDS - INFO - Fetching effects for (93, '2014-11-10 13:11:02', 'hst_0297.pmap', 'The WFC3 Team delivered 3 new dark reference files.')
    ####################################################################################################
    --------------------------------------------------------------------------------------------------------------
    CRDS hst ops datasets affected hst_0295.pmap --> hst_0296.pmap on 2014-11-05-13:55:08
    --------------------------------------------------------------------------------------------------------------
    CRDS - INFO - [2014-11-05 13:55:12,922]  Mapping differences from 'hst_0295.pmap' --> 'hst_0296.pmap' affect:
     {'acs': ['biasfile', 'drkcfile', 'darkfile']}
    CRDS - INFO - [2014-11-05 13:55:13,576]  Possibly affected --datasets-since dates determined by 'hst_0295.pmap' --> 'hst_0296.pmap' are:
     {'acs': '2014-08-26 09:55:53'}
    CRDS - INFO - [2014-11-05 13:55:13,576]  Computing bestrefs for db datasets for ['acs']
    CRDS - INFO - [2014-11-05 13:55:13,576]  Dumping dataset parameters for 'acs' from CRDS server at 'https://hst-crds.stsci.edu' since '2014-08-26 09:55:53'
    CRDS - INFO - [2014-11-05 13:55:39,323]  Downloaded  1431 dataset ids for 'acs' since '2014-08-26 09:55:53'
    CRDS - INFO - [2014-11-05 13:55:52,999]  Updated exposure counts:
     {'ACS': {'biasfile': 1374, 'darkfile': 1386, 'drkcfile': 1386}}
    CRDS - INFO - [2014-11-05 13:55:53,001]  Affected products = 670
    CRDS - INFO - [2014-11-05 13:55:53,001]  Unique error types: 0
    CRDS - INFO - [2014-11-05 13:55:53,001]  STARTED 2014-11-05 13:55:10.62
    CRDS - INFO - [2014-11-05 13:55:53,002]  STOPPED 2014-11-05 13:55:53.00
    CRDS - INFO - [2014-11-05 13:55:53,002]  ELAPSED 0:00:42.37
    CRDS - INFO - [2014-11-05 13:55:53,002]  1.43 K datasets at 33.77  datasets-per-second
    CRDS - INFO - [2014-11-05 13:55:53,002]  0 errors
    CRDS - INFO - [2014-11-05 13:55:53,002]  0 warnings
    CRDS - INFO - [2014-11-05 13:55:53,002]  12 infos
    --------------------------------------------------------------------------------------------------------------
    OK: CRDS hst ops datasets affected hst_0295.pmap --> hst_0296.pmap on 2014-11-05-13:55:08 : 670 affected
    --------------------------------------------------------------------------------------------------------------
    ####################################################################################################
    --------------------------------------------------------------------------------------------------------------
    CRDS hst ops datasets affected hst_0296.pmap --> hst_0297.pmap on 2014-11-10-13:15:07
    --------------------------------------------------------------------------------------------------------------
    CRDS - INFO - [2014-11-10 13:15:10,393]  Mapping differences from 'hst_0296.pmap' --> 'hst_0297.pmap' affect:
     {'wfc3': ['darkfile']}
    CRDS - INFO - [2014-11-10 13:15:10,574]  Possibly affected --datasets-since dates determined by 'hst_0296.pmap' --> 'hst_0297.pmap' are:
     {'wfc3': '2014-10-27 00:30:40'}
    CRDS - INFO - [2014-11-10 13:15:10,574]  Computing bestrefs for db datasets for ['wfc3']
    CRDS - INFO - [2014-11-10 13:15:10,574]  Dumping dataset parameters for 'wfc3' from CRDS server at 'https://hst-crds.stsci.edu' since '2014-10-27 00:30:40'
    CRDS - INFO - [2014-11-10 13:15:33,868]  Downloaded  714 dataset ids for 'wfc3' since '2014-10-27 00:30:40'
    CRDS - INFO - [2014-11-10 13:15:43,290]  Updated exposure counts:
     {'WFC3': {'darkfile': 369}}
    CRDS - INFO - [2014-11-10 13:15:43,291]  Affected products = 292
    CRDS - INFO - [2014-11-10 13:15:43,291]  Unique error types: 0
    CRDS - INFO - [2014-11-10 13:15:43,291]  STARTED 2014-11-10 13:15:08.58
    CRDS - INFO - [2014-11-10 13:15:43,291]  STOPPED 2014-11-10 13:15:43.29
    CRDS - INFO - [2014-11-10 13:15:43,291]  ELAPSED 0:00:34.70
    CRDS - INFO - [2014-11-10 13:15:43,291]  714 datasets at 20.57  datasets-per-second
    CRDS - INFO - [2014-11-10 13:15:43,292]  0 errors
    CRDS - INFO - [2014-11-10 13:15:43,292]  0 warnings
    CRDS - INFO - [2014-11-10 13:15:43,292]  12 infos
    --------------------------------------------------------------------------------------------------------------
    OK: CRDS hst ops datasets affected hst_0296.pmap --> hst_0297.pmap on 2014-11-10-13:15:07 : 292 affected
    --------------------------------------------------------------------------------------------------------------
    ####################################################################################################
    CRDS - INFO - Contributing context switches = 2
    CRDS - INFO - Total products affected = 962
    CRDS - INFO - 0 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 4 infos

QAD is designed to be run perdiodically,  say every 30 minutes,  to check with the CRDS server for
context updates.  If multiple context switches occur during one polling interval,  by default QAD
includes IDs from all of them.   This is also true of interactive queries using -x and/or -y,  so
it's possible to combine affected datasets from multiple switches when repeated IDs are expected.

QAD also supports a --single-context-switch (-s) mode for printing results 1-by-1 in the advent of
multiple context switches in on polling interval.

Since there is a race condition between when a context is made operational and when affected datasets results
are available on the CRDS server,  QAD also supports a -i switch for ignoring unavailable results.  Alternately
missing results are considered an error,  the main difference being the exit status.

For initialzing,  specify -i to ignore any missing computations since QAD will attempt to process the
entire history the first time it is run and precomputed results don't exist for all historical context
switches.

It's possible for precomputed results to contain bestrefs errors of some sort,  most likely due to invalid
bestrefs selection parametersin the HST DADSOPS catalog.   By default the datasets from a computation which
contained errors are excluded from the overall results.   Use -k to include the dataset IDs from computations
which included errors.

Conversely, to abort processing when an affected datasets computation included errors,  use -z to fail
and quit.

The --quiet (-q) parameter suppresses the recorded log output from the affected datasets computations:

    % query_affected_datasets -x 94 -y 97 -q > ids
    CRDS - INFO - Fetching effects for (94, '2014-11-18 17:15:35', 'hst_0298.pmap', 'Delivery of new ACS DKC, DRK, and BIA files.')
    CRDS - INFO - Fetching effects for (95, '2014-11-20 16:12:34', 'hst_0299.pmap', 'Delivery of new WFC3 UVIS darks.')
    CRDS - INFO - Fetching effects for (96, '2014-11-25 15:48:40', 'hst_0300.pmap', 'Delivery of a new COS HVTAB for association LCIX02080.')
    CRDS - INFO - Contributing context switches = 3
    CRDS - INFO - Total products affected = 10356
    CRDS - INFO - 0 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 5 infos

NOTE:  CRDS logging is used in both query_affected_datasets and the original server-side affected datasets computations.  The
final errors count shown above only applies to the client-side computing in query_affected_datasets,  so server-side errors are
not *counted*.   However,  server-side errors are tracked and reduced to a single client-side error for each server-side
bestrefs run with errors.

The --verbose parameter includes debug output in excess of normal application logging,  possibly useful
for debugging subclasses of the QueryAffectedDatasetsScript skeletal framework.

"""
    def add_args(self):
        """Add diff-specific command line parameters."""

        self.add_argument("-l", "--list-history", dest="list_history", action="store_true",
            help="Print out the context history and exit.")

        self.add_argument("-x", "--starting-context", dest="starting_context",
            help="Use the affected datasets computation starting with this history index integer, date, or context name. Defaults to last processed.",
            metavar="INT_OR_NAME", default=None, type=str)

        self.add_argument("-y", "--stopping-context", dest="stopping_context",
            help="Use the affected datasets computation starting with this history index integer, date, or context name. Defaults to end of history.",
            metavar="INT_OR_NAME", default=None, type=str)

        self.add_argument("-s", "--single-context-switch", dest="single_context_switch", action="store_true",
            help="For default indexing, if multiple new contexts are available,  just process one new context and stop.")

        self.add_argument("-i", "--ignore-missing-results", dest="ignore_missing_results", action="store_true",
            help="Skip over any requested context switch which has no pre-computed results on the CRDS server.  Otherwise fatal.")

        self.add_argument("-k", "--ignore-errant-history", dest="ignore_errant_history", action="store_true",
            help="If bestrefs status indicates errors occurred, issue an error message but include the dataset ids in results.")

        self.add_argument("-z", "--fail-on-errant-history", dest="fail_on_errant_history", action="store_true",
            help="If bestrefs status indicates errors occurred, quit processing.  (fix the server and rerun).")

        self.add_argument("-f", "--last-processed-file", dest="last_processed_file",
            help="File containing the tuple of the last history item successfully processed. Defaults to file in CRDS cache.")

        self.add_argument("-q", "--quiet", dest="quiet", action="store_true",
            help="Terser log output.")

        self.add_argument("-r", "--reset", dest="reset", action="store_true",
            help="Reset the last-context-processed file to the end of the current history.  Useful for init and reinit.")

    def __init__(self, *args, **keys):
        super(QueryAffectedDatasetsScript, self).__init__(*args, **keys)
        self.contributing_context_switches = 0

    def main(self):
        """Top level processing method."""
        self.require_server_connection()
        if self.args.list_history:
            return self.list_history()
        if self.args.reset:
            return self.reset_last_processed()
        effects = self.polled()
        ids = self.process(effects)
        self.use_all_ids(effects, ids)
        self.log_all_ids(effects, ids)
        if effects:
            self.save_last_processed(effects)
        return log.errors()

    def process(self,  effects):
        """Output the results of all the context transitions."""
        ids = []
        for i, affected in effects:
            self.log_affected(i, affected)
            if not self.ignore_errors(i, affected):
                ids.extend(affected.affected_ids)
                self.use_affected(i, affected)
                self.contributing_context_switches += 1
        return sorted(set(ids))

    def use_affected(self, i, affected):
        """PLUGIN: for doing something with each individual context switch set of effects. Default does nothing.
        i         -- the ending history index of the transition
        affected  -- dictionary of effects info
        """
        pass

    def use_all_ids(self, effects, ids):
        """PLUGIN: for using all ids which passed availability and error screening. Default prints ids to stdout.

        effects :    [ (history_index, affects info), ...]

        ids :  sorted set of all ids affected by specified or implied contexts
        """
        if ids:
            print("\n".join(ids))

    def log_affected(self, i, affected):
        """PLUGIN: Banner log and debug output for each context switch."""
        if log.get_verbose():
            print("#"*100, file=sys.stderr)
            log.debug("History:", i, "Effects:\n", log.PP(affected))
        else:
            if not self.args.quiet:
                print("#"*100, file=sys.stderr)
                print(affected.bestrefs_err_summary, file=sys.stderr)

    def log_all_ids(self, effects, ids):
        """PLUGIN: Summary output after all contexts processed."""
        if self.args.quiet:
            return
        if not effects:
            log.info("No new results are available.")
        else:
            if not ids:
                log.info("No ids were affected.")
            print("#"*100, file=sys.stderr)
            log.info("Contributing context switches =", len(effects))
            log.info("Total products affected =", len(ids))
        log.standard_status()

    def list_history(self):
        """Print out the context history."""
        for i in range(self.history_start, self.history_stop+1):
            print((i,) + self.history[i])

    @property
    @utils.cached
    def history(self):
        """Return the context history or fail.  Should nominally always work."""
        try:
            return api.get_context_history(self.observatory)
        except Exception as exc:
            self.fatal_error("get_context_history failed: ", str(exc).replace("OtherError:",""))

    def get_affected(self, old_context, new_context):
        """Return the affected datasets Struct for the transition from old_context to new_context,
        or None if the results aren't ready yet.
        """
        try:
            affected = api.get_affected_datasets(self.observatory, old_context, new_context)
        except Exception as exc:
            if "No precomputed affected datasets results exist" in str(exc):
                if self.args.ignore_missing_results:
                    log.info("No results for", old_context, "-->", new_context, "ignoring and proceeding.")
                    affected = None
                else:
                    self.fatal_error("Results for", old_context, "-->", new_context, "don't exist or are not yet complete.")
            else:
                self.fatal_error("get_affected_datasets failed: ", str(exc).replace("OtherError:",""))
        return affected

    def ignore_errors(self, i, affected):
        """Check each context switch for errors during bestrefs run. Fail or return False on errors."""
        ignore = False
        if affected.bestrefs_status != 0:
            message = log.format("CRDS server-side errors for", i, affected.computation_dir)
            if self.args.ignore_errant_history:
                ignore = True
            if self.args.fail_on_errant_history:
                self.fatal_error(message)
            else:
                log.error(message)
        return ignore

    def polled(self):
        """Output the latest affected datasets taken from the history starting item onward.
        Since the history drives and ultimately precedes any affected datasets computation,  there's
        no guarantee that every history item is available.
        """
        assert 0 <= self.history_start < len(self.history), "Invalid history interval with starting index " + repr(self.history_start)
        assert 0 <= self.history_stop  < len(self.history), "Invalid history interval with stopping index " + repr(self.history_stop)
        assert self.history_start <= self.history_stop, "Invalid history interval,  start >= stop."
        effects = []
        for i in range(self.history_start, self.history_stop):
            old_context = self.history[i][1]
            new_context = self.history[i+1][1]
            if old_context > new_context:   # skip over backward transitions,  no output.
                continue
            log.info("Fetching effects for", (i,) + self.history[i+1])
            affected = self.get_affected(old_context, new_context)
            if affected:
                effects.append((i, affected))
        return effects

    @property
    @utils.cached
    def history_start(self):
        """Return the starting item for processing."""
        if self.args.starting_context:
            item = self.convert_context(self.args.starting_context)
        else:  # read the start from the file recording the last successful processing index.
            if os.path.exists(self.last_processed_path) and not self.args.list_history:
                try:
                    last_tuple = utils.evalfile(self.last_processed_path)
                    item = int(last_tuple[0])
                    log.verbose("Loaded last processed:", last_tuple)
                except Exception as exc:
                    self.fatal_error(self.last_processed_path, "already exists but was unusable:", str(exc))
            else:
                item = 0
        return item

    @property
    @utils.cached
    def history_stop(self):
        """Return the stopping item for processing."""
        if self.args.stopping_context:
            item = self.convert_context(self.args.stopping_context)
        else:  # read the start from the file recording the last successful processing index.
            if self.args.single_context_switch:
                item = self.history_start + 1
                if item >= len(self.history) - 1:
                    item = len(self.history) - 1
            else:
                item = len(self.history) - 1
        return item

    def convert_context(self, context):
        """Convert an integer or context name into a history index."""
        try:
            return int(context)
        except ValueError:
            pass
        try:
            date = timestamp.reformat_date(context)
            is_date = True
        except:
            is_date = False
        hist = self.history
        for item, hist in enumerate(hist):
            if is_date and hist[0] >= date:
                break
            elif context in hist[1]:
                break
        else:
            self.fatal_error("Context = '{}' not found in history".format(context))
        assert 0 <= item < len(self.history),  "Invalid history item " + repr(item)
        return item

    def save_last_processed(self, effects):
        """Record the last index history tuple successfully processed."""
        last_ix = effects[-1][0]
        if last_ix == -1:
            return
        hist = (last_ix+1,) + tuple(self.history[last_ix + 1])
        self._write_last_processed(hist)

    def reset_last_processed(self):
        """Reset the state of the last context processed marker to the end of the current history."""
        hist = (len(self.history)-1,) + tuple(self.history[-1])
        self._write_last_processed(hist)

    def _write_last_processed(self, hist):
        """Write down the history tuple of the last context processed."""
        log.verbose("Saving last processed:", repr(hist))
        log.verbose("Storing last processed state at", repr(self.last_processed_path))
        utils.ensure_dir_exists(self.last_processed_path)
        with open(self.last_processed_path, "w+") as last:
            last.write(str(hist))

    @property
    def last_processed_path(self):
        """Path of file recording last processed context."""
        if self.args.last_processed_file:
            return self.args.last_processed_file
        else:
            return os.path.join(config.get_crds_cfgpath(self.observatory), "ad_last_processed")

if __name__ == "__main__":
    sys.exit(QueryAffectedDatasetsScript()())



'''This module adds additional submission metadata and a programmatic interface
to the original commamnd line submission program as part of the file submission
streamlining project.'''

import os
import sys
import yaml
import collections
import webbrowser

from crds.core import log, config
from .submit import ReferenceSubmissionScript
import urllib
from functools import wraps
from textwrap import wrap

class NoFilesSelected(Exception):
    pass

BASE_URLS = {
    'dev': {
        'hst':  'https://hst-crds-dev.stsci.edu/',
        'jwst': 'https://jwst-crds-dev.stsci.edu/',
        'roman': 'https://roman-crds-dev.stsci.edu/',
    },
    'test': {
        'hst':  'https://hst-crds-test.stsci.edu/',
        'jwst': 'https://jwst-crds-test.stsci.edu/',
        'roman': 'https://roman-crds-test.stsci.edu/',
    },
    'ops': {
        'hst':  'https://hst-crds.stsci.edu/',   # 'ops' not in URL
        'jwst': 'https://jwst-crds.stsci.edu/',
        'roman': 'https://roman-crds.stsci.edu/',
    },
    'bit': {
        'hst':  'https://hst-crds-bit.stsci.edu/',
        'jwst': 'https://jwst-crds-bit.stsci.edu/',
    },
    'cit': {
        'hst':  'https://hst-crds-cit.stsci.edu/',
        'jwst': 'https://jwst-crds-cit.stsci.edu/',
    },
    'int': {
        'roman': 'https://roman-crds-int.stsci.edu/',
    },
}

URL_DESCRIPTION = 'submission_form/redcat_description.yml'

NULL_FIELDTYPES = {
    'BooleanField'     : bool,
    'CharField'        : str,
    'TypedChoiceField' : str, }

# Preserve order of YAML dicts (from https://stackoverflow.com/a/52621703):
yaml.add_representer(dict, lambda self, data: yaml.representer.SafeRepresenter.represent_dict(self, data.items()))

class RedCatApiScript(ReferenceSubmissionScript):
    '''This script extends the original CRDS command line submission
    system with extra meta-data previously captured by the submit_to_redcat
    script.'''
    def __init__(self, *args, **kargs):
        self._extra_redcat_parameters = {}
        super(RedCatApiScript, self).__init__(*args, **kargs)

    def get_submission_args(self):
        '''Returns the combined form parameter dictionary as a CRDS Struct defining
        the mapping from all form variables to their string values.'''
        parameters = super(RedCatApiScript, self).get_submission_args()
        extra_parameters = self.get_extra_parameters()
        parameters.update(extra_parameters)
        return parameters

    def get_extra_parameters(self):
        '''Return the form dictionary mapping form variables to value strings for
        new variables being added by the streamlining project.'''
        return self._extra_redcat_parameters

    def batch_submit_references(self):
        '''Do a web re-post to the batch submit references web page.'''
        return self._submission("/submission_form/redcat_submit/")

class RedCatSubmissionScript(RedCatApiScript):
    '''This script extends the original CRDS command line submission
    system with extra meta-data previously captured by the submit_to_redcat
    script.'''
    def add_args(self):
        super(RedCatSubmissionScript, self).add_args()
        self.add_argument("--redcat-parameters", type=str,
                     help="Path to YAML file defining extra ReDCaT parameters.")

    def get_extra_parameters(self):
        '''Return the form dictionary mapping form variables to value strings for
        new variables being added by the streamlining project.'''
        with open(self.args.redcat_parameters) as f:
            text = f.read()
            log.verbose("Raw YAML read:\n", text, verbosity=75)
            loaded = yaml.safe_load(text)
            log.verbose("ReDCaT parameters:\n", log.PP(loaded))
            return loaded

class SubmissionResult(collections.namedtuple("SubmissionResult", ["error_count", "warning_count", "ready_url"])):
    def open_ready_url(self):
        if self.ready_url is None:
            raise RuntimeError("ready_url is not present")
        webbrowser.open(self.ready_url)

class Submission(object):
    '''Client-side Redcat submission class.  Can be used to prepare, validate, and submit
    CRDS submissions.

    Call `submission_obj.help()` to print details about the submission object form fields.

    Parameters:
        observatory (str, {hst, jwst}):  Used in determining which CRDS for submission
        string (str, {ops, test, dev, bit, cit}):  Used in determining which CRDS for submission
        context (str): Derived-from context.  If omitted, will default to the observatory's edit context.'''
    def __init__(self, observatory, string, context=None, *args, **kwargs):
        observatory, string = observatory.lower(), string.lower()
        if observatory not in ['hst', 'jwst', 'roman']:
            raise ValueError('Observatory "{}" must be one of "hst" or "jwst" or "roman".')
        if string not in ['dev', 'test', 'ops', 'bit', 'cit', 'int']:
            raise ValueError('String "{}" must be either "ops", "test", "dev", "bit", "cit", or "int".')
        self._observatory = observatory
        self._string = string
        self._context = context

        config.base_url = BASE_URLS[self.string][self.observatory]
        if not os.environ.get('CRDS_SERVER_URL'):
            log.verbose('Setting $CRDS_SERVER_URL to {}'.format(config.base_url))
            os.environ['CRDS_SERVER_URL'] = config.base_url

        url = urllib.parse.urljoin(config.base_url, URL_DESCRIPTION)
        try:
            with urllib.request.urlopen(url) as req:
                self._form_description = yaml.safe_load(req)
        except (urllib.error.HTTPError, urllib.error.URLError, ) as e:
            print ('Check your network connection!')
            raise e
        # Convert list describing form to a dictionary (preserves order):
        self._form_description = {field['key']: field for field in self._form_description}
        for key in self._form_description:  self._form_description[key].pop('key')

        # Hard-code in change_level and description for now:
        self._form_description['change_level'] = {
            'type': 'TypedChoiceField',
            'choices': {'TRIVIAL', 'MODERATE', 'SEVERE'},
            'initial': 'SEVERE',
            'required': True,
            'label': 'Degree that new files are expected to impact science results.', }
        self._form_description['description'] = {
            'type': 'CharField',
            'required': True,
            'label': 'Information about file changes and expected impacts, include '
                     'instrument and type.  Formatting note: only alphanumeric, '
                     'periods, commas, dashes, and underscores are allowed', }

        self._all_keys = set(self._form_description.keys())
        self._required_keys = {x for x in self._form_description if self._form_description[x]['required']}
        self._optional_keys = {x for x in self._form_description if not self._form_description[x]['required']}

        self._fields = dict()  # Users should not modify this dict directly!
        for key in self._form_description:
            # Avoid field validation for initialization by accessing hidden dictionary directly:
            self._fields[key] = NULL_FIELDTYPES[self._form_description[key]['type']]()
            try:
                self[key] = self._form_description[key]['initial']
            except KeyError:
                pass

        self._files = set()  # Users should not modify this directly!

    def __repr__(self):
        return '<Submission Object {}-{}>:\nFields:  {}\nFiles:  {}'.format(
            self.observatory, self.string,
            self._fields.__repr__(),
            list(self.files))

    # ------------------------------------------------------------------------------------
    # Pass some dictionary methods to handle form field manipulation with prototyping:

    def __setitem__(self, key, value):
        '''Intercept and enforce validation requirements on individual fields.
        Booleans values map to 'Yes' and 'No' str for character fields.'''
        if key not in self._all_keys:
            raise KeyError("Key not in submission form template:  '{}'".format(key))
        if (key in self._required_keys) and (value == ''):  # allow Boolean False
            raise ValueError("Field '{}' cannot be empty.".format(key))
        field_type = NULL_FIELDTYPES[self._form_description[key]['type']]

        # Interpret boolean values in choice fields as 'Yes' and 'No':
        if isinstance(value, bool) and ('choices' in self._form_description[key]) and \
            not (self._form_description[key]['type'] == 'BooleanField'):
            if value:
                value = 'Yes'
            else:
                value = 'No'
        if not isinstance(value, field_type):
            raise ValueError("'{}' must be of type {}".format(key, field_type.__name__))

        # Check if choice fields have allowed values:
        if ('choices' in self._form_description[key]) and \
           (self._form_description[key]['type'] != 'BooleanField'):
            matches = [x for x in self._form_description[key]['choices'] if x.lower() == value.lower()]
            if len(matches) != 1:
                raise ValueError("'{}' must be a valid choice: {{{}}}".format(key,
                                 ', '.join(self._form_description[key]['choices'])))
            # Inherit case from matching choice:
            value = matches[0]

        self._fields[key] = value

    def __delitem__(self, key, *args, **kargs):
        '''Reset self[key] to its default initialized value.'''
        if key in self:
            self._fields[key] = NULL_FIELDTYPES[self._form_description[key]['type']]()
            try:
                self[key] = self._form_description[key]['initial']
            except KeyError:
                pass
        else:
            raise KeyError(key)

    @wraps(dict.__getitem__)
    def __getitem__(self, key, *args, **kargs):
        return self._fields.__getitem__(key, *args, **kargs)

    @wraps(dict.__contains__)
    def __contains__(self, *args, **kargs):
        return self._fields.__contains__(*args, **kargs)

    @wraps(dict.get)
    def get(self, *args, **kargs):
        return self._fields.get(*args, **kargs)

    @wraps(dict.keys)
    def keys(self, *args, **kargs):
        return self._fields.keys(*args, **kargs)

    @wraps(dict.values)
    def values(self, *args, **kargs):
        return self._fields.values(*args, **kargs)

    @wraps(dict.items)
    def items(self, *args, **kargs):
        return self._fields.items(*args, **kargs)

    # ------------------------------------------------------------------------------------
    # Set methods to handle filename manipulation:

    def add_file(self, filename):
        '''Add a file to the submission.  Calls crds.certify() on the file.'''
        if not os.access(filename, os.R_OK):
            raise FileNotFoundError("'{}' does not exist or is not readable.".format(filename))

        self._files.add(filename)

    @wraps(set.remove)
    def remove_file(self, filename, *args, **kargs):
        self._files.remove(filename, *args, **kargs)

    # ------------------------------------------------------------------------------------
    # Class properties protected from direct user manipulation:

    @property
    def observatory(self):
        '''Instantiated for HST or JWST.'''
        return self._observatory

    @property
    def string(self):
        '''Instantiated for ops, test, dev, bit, or cit string.'''
        return self._string

    @property
    def context(self):
        '''User specified derived-from context for the submission.'''
        return self._context

    @property
    def files(self):
        '''Set of files associated with the submission.'''
        return list(self._files)

    @property
    def base_url(self):
        return config.base_url

    # ------------------------------------------------------------------------------------
    # Custom methods:

    def help(self):
        '''Print help text derived from CRDS instance specified.'''
        # Can't easily overwrite __doc__ dynamically.
        for key, field in self._form_description.items():
            print (key, ' (', NULL_FIELDTYPES[self._form_description[key]['type']].__name__,
                ', optional)' if not field.get('required', False) else ')', '\n', '-'*len(key), sep='')
            print ('\n'.join(wrap(field['label'])))
            if 'help_text' in field:
                print ('\n'.join(wrap(field['help_text'])))
            if 'choices' in field:
                print ('Valid choices:')
                print ('  {', ', '.join(["'{}'".format(x) for x in field['choices']]), '}', sep='')
            print ()

    def validate(self):
        '''Validate the object for submission to CRDS.'''
        if (set(self.keys()) - self._optional_keys) != self._required_keys:
            raise Exception('Extra/missing keys')

        # Check for all empty required keys at once to raise one exception:
        empty_keys = {key for key in self._required_keys if self[key] == ''}  # Don't flag False booleans
        if empty_keys:
            raise ValueError('These keywords cannot be empty:\n    ' + '\n    '.join(empty_keys))

        # Make sure files were associated with the submission:
        if len(self.files) == 0:
            raise NoFilesSelected('No files have been added to submission.  '
                                  'Use the `submission_obj.add_file()` method.')

    @wraps(yaml.safe_dump)
    def yaml(self, *args, **kargs):
        '''YAML representation of this submission object.'''
        return yaml.safe_dump(dict(self), *args, **kargs)

    def submit(self):
        '''Validate submission form, upload to CRDS staging, handle server-side
        submission errors.'''
        # Client-side validation
        # Upload the form
        # Handle returned server-side errors

        self.validate()
        argv = [ "crds.submit", "--files" ] + self.files + [
            "--monitor-processing",
            "--wait-for-completion",
            "--wipe-existing-files",
            "--certify-files",
            "--log-time",
            "--stats",
            "--creator", "{} Team".format(self['instrument']),
            "--change-level", self["change_level"],
            "--description", self["description"],
            ]

        if self.context is not None:
            argv.extend(["--derive-from-context", self.context])

        log.verbose(argv)
        script = RedCatApiScript(argv)
        script._extra_redcat_parameters = dict(self)
        script()

        return SubmissionResult(
            error_count=script._error_count,
            warning_count=script._warning_count,
            ready_url=script._ready_url
        )

def main():
    '''Run the command line program version of the extended batch submit which
    loads extended parameters from a YAML input file.'''
    script = RedCatSubmissionScript()
    return script()

if __name__ == '__main__':
    sys.exit(main())



import os.path

from crds.core import reftypes

HERE  = os.path.dirname(__file__) or "."

TYPES = reftypes.from_package_file("tobs", __file__)

OBSERVATORY = TYPES.observatory
INSTRUMENTS = TYPES.instruments
EXTENSIONS = TYPES.extensions
TEXT_DESCR = TYPES.text_descr
FILEKINDS = TYPES.filekinds

INSTRUMENT_FIXERS = {
}

TYPE_FIXERS = {
}

PROVENANCE_KEYWORDS = ("META.DESCRIPTION", "META.PEDIGREE", "META.USEAFTER","META.HISTORY", "META.AUTHOR")
# PROVENANCE_KEYWORDS = ("DESCRIP", "PEDIGREE", "USEAFTER","HISTORY", "AUTHOR")
USEAFTER_KEYWORDS = ("META.OBSERVATION.DATE", "META.OBSERVATION.TIME") # Dataset keywords matching in UseAfter selectors

DEFAULT_SELECTORS = ("Match", "UseAfter")   # Normal selector hierarchy in rmap



'''
This module defines top level functions for operating on files,  either related to
determining fundamental properties of a file (get_filetype, get_observatory) or
related to instantiating a specialized file handler object (file_factory) appropriate
for the given observatory and file format.
'''

import os

# ============================================================================

from astropy.io import fits as pyfits

# ============================================================================

# Defer these to avoid required heavy weight dependencies like "jwst" or "asdf"
# from crds.io import fits, json, yaml, geis, asdf, jwstdm

# ============================================================================

from crds.core import config, utils, constants

# ============================================================================

def file_factory(filepath, original_name=None, observatory=None):
    """Based on parameters,  construct a file object of the appropriate
    class.  Unspecified parameters determined from `filepath` as possible.

    Where possible file extension will be used to determine file type.

    filepath        must be a viable path to an existing reference
    original_name   may be a more informative basename if filepath is a temp file
    observatory     "hst" or "jwst",  biases JWST to use datamodels vs. fits
    """
    filetype = get_filetype(filepath, original_name)
    if filetype == "asdf":
        from crds.io import asdf
        file_class = asdf.AsdfFile
    elif filetype == "json":
        from crds.io import json
        file_class = json.JsonFile
    elif filetype == "yaml":
        from crds.io import yaml
        file_class = yaml.YamlFile
    elif filetype == "geis":
        from crds.io import geis
        file_class = geis.GeisFile
    elif filetype == "fits":
        from crds.io import fits
        file_class = fits.FitsFile
    else:
        raise RuntimeError("Unknown file type for " + repr(filepath) )
    if observatory is None:
        observatory = get_observatory(filepath, original_name) # slow?
    return file_class(filepath, original_name, observatory)

# ----------------------------------------------------------------------------------------------

@utils.gc_collected
def get_observatory(filepath, original_name=None):
    """Return the observatory corresponding to `filepath`.  filepath
    may be a web temporary file with a garbage name.   Use
    `original_name` to make inferences based on file extension, or
    filepath if original_name is None.
    """
    if original_name is None:
        original_name = filepath
    for observatory in constants.ALL_OBSERVATORIES:
        if original_name.startswith(observatory + "_"):
            return observatory
    observatory = "hst"
    if original_name.endswith(".fits"):
        try:
            observatory = pyfits.getval(filepath, keyword="TELESCOP")
        except KeyError:
            pass
    elif original_name.endswith(".asdf"):
        try:
            import asdf
            with asdf.open(filepath) as handle:
                observatory = handle["meta"]["telescope"]
        except KeyError:
            pass
    elif original_name.endswith((".yaml", ".json", ".text", ".txt")):
        return "jwst"
    return observatory.lower()

# ----------------------------------------------------------------------------------------------

def get_filetype(filepath, original_name=None):
    """Determine file type from `original_name` if possible, otherwise attempt to
    idenitfy based on file contents.
    """
    if original_name is None:
        original_name = os.path.basename(filepath)

    # Fast extension-based type determination
    filetype = config.filetype(original_name)
    if filetype != "unknown":
        return filetype

    with open(filepath, "rb") as handle:
        first_5 = str(handle.read(5).decode('utf-8'))
        if first_5 == "#ASDF":
            return "asdf"
        elif first_5 == "SIMPL":
            first_81 = first_5 + handle.read(76).decode('utf-8')
            if first_81[-1] == '\n':
                all_chars = first_81 + handle.read().decode('utf-8')
                lines = all_chars.splitlines()
                lengths = { len(line) for line in lines }
                if len(lengths) == 1 and 80 in lengths:
                    return 'geis'
                else:
                    return 'fits'
            else:
                return "fits"

    try:
        with open(filepath) as handle:
            import json
            json.load(handle)
            return "json"
    except Exception:
        pass

    try:
        with open(filepath) as handle:
            import yaml
            yaml.safe_load(handle)
            return "yaml"
    except Exception:
        pass

    return "unknown"

# ----------------------------------------------------------------------------------------------

def is_dataset(name):
    """Returns True IFF `name` is plausible as a dataset.   Not a guarantee."""
    return config.filetype(name) in ["fits", "asdf", "geis"]
